import json
import os
from typing import Any, Dict, Optional, Union

import aiohttp

from ...exceptions import (
    CollectionNotFoundError,
    DownloadFilesError,
    GetTaskError,
    UploadArtifactsError,
    UploadRequestsError,
)
from ...helper.decorators import require_api_key


class DataOperationsMixin:
    """Data generation and processing operations."""
    
    @require_api_key
    async def upload_artifacts(
        self,
        collection: str,
        file_type: str,
        compressed: Optional[bool] = True
    ) -> Dict[str, Any]:
        """
        Retrieve signed url to upload artifact file to a collection.

        Args:
            collection: Name of collection
            file_type: The extension of the file
            compressed: Whether to compress the file using gzip or not (defaults to True)
        
        Returns:
            API response as a dictionary containing the upload URL

        Raises:
            CollectionNotFoundError: If the collection is not found
            UploadArtifactsError: If the API request fails due to unknown reasons
        """
        params = {
            "file_type": file_type,
            "compressed": str(compressed).lower(),
        }

        response, status = await self._client._terrakio_request("GET", f"collections/{collection}/upload", params=params)

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise UploadArtifactsError(f"Upload artifacts failed with status {status}", status_code=status)

        return response

    async def _get_upload_url(
        self,
        collection: str
    ) -> Dict[str, Any]:
        """
        Retrieve signed url to upload requests for a collection.

        Args:
            collection: Name of collection
        
        Returns:
            API response as a dictionary containing the upload URL

        Raises:
            CollectionNotFoundError: If the collection is not found
            UploadRequestsError: If the API request fails due to unknown reasons
        """
        response, status = await self._client._terrakio_request("GET", f"collections/{collection}/upload/requests")

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise UploadRequestsError(f"Upload requests failed with status {status}", status_code=status)

        return response

    @require_api_key
    async def _upload_file(self, file_path: str, url: str, use_gzip: bool = True):
        """
        Helper method to upload a JSON file to a signed URL.
        
        Args:
            file_path: Path to the JSON file
            url: Signed URL to upload to
            use_gzip: Whether to compress the file with gzip
        """
        try:
            with open(file_path, 'r') as file:
                json_data = json.load(file)
        except FileNotFoundError:
            raise FileNotFoundError(f"JSON file not found: {file_path}")
        except json.JSONDecodeError as e:
            raise ValueError(f"Invalid JSON in file {file_path}: {e}")
        
        return await self._upload_json_data(json_data, url, use_gzip)

    @require_api_key
    async def _upload_json_data(self, json_data, url: str, use_gzip: bool = True):
        """
        Helper method to upload JSON data directly to a signed URL.
        
        Args:
            json_data: JSON data (dict or list) to upload
            url: Signed URL to upload to
            use_gzip: Whether to compress the data with gzip
        """
        if hasattr(json, 'dumps') and 'ignore_nan' in json.dumps.__code__.co_varnames:
            dumps_kwargs = {'ignore_nan': True}
        else:
            dumps_kwargs = {}
        
        if use_gzip:
            import gzip
            body = gzip.compress(json.dumps(json_data, **dumps_kwargs).encode('utf-8'))
            headers = {
                'Content-Type': 'application/json',
                'Content-Encoding': 'gzip'
            }
        else:
            body = json.dumps(json_data, **dumps_kwargs).encode('utf-8')
            headers = {
                'Content-Type': 'application/json'
            }
        
        response = await self._client._regular_request("PUT", url, data=body, headers=headers)
        return response

    @require_api_key
    async def generate_data(
        self,
        collection: str,
        file_path: str,
        output: str,
        skip_existing: Optional[bool] = True,
        force_loc: Optional[bool] = None,
        server: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Generate data for a collection.

        Args:
            collection: Name of collection
            file_path: Path to the file to upload
            output: Output type (str)
            force_loc: Write data directly to the cloud under this folder
            skip_existing: Skip existing data
            server: Server to use
        
        Returns:
            API response as a dictionary containing task information

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails due to unknown reasons
        """
        
        await self.get_collection(collection = collection)

        upload_urls = await self._get_upload_url(
            collection = collection
        )
        
        url = upload_urls['url']

        await self._upload_file(file_path, url)
        
        payload = {"output": output, "skip_existing": skip_existing}
        
        if force_loc is not None:
            payload["force_loc"] = force_loc
        if server is not None:
            payload["server"] = server
        
        response, status = await self._client._terrakio_request("POST", f"collections/{collection}/generate_data", json=payload)

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise GetTaskError(f"Generate data failed with status {status}", status_code=status)
        
        return response

    @require_api_key
    async def post_processing(
        self,
        collection: str,
        folder: str,
        consumer: str,
        extra: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Run post processing for a collection.
        
        Args:
            collection: Name of collection
            folder: Folder to store output
            consumer: Path to post processing script
            extra: Optional JSON string for extra params (dict)
        
        Returns:
            API response as a dictionary containing task information
        
        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails due to unknown reasons
        """
        await self.get_collection(collection=collection)
        
        with open(consumer, 'rb') as f:
            form = aiohttp.FormData()
            form.add_field('folder', folder)
            form.add_field(
                'consumer',
                f.read(),
                filename='consumer.py',
                content_type='text/x-python'
            )
            if extra is not None:
                form.add_field('extra', extra)
            
            response, status = await self._client._terrakio_request(
                "POST",
                f"collections/{collection}/post_process",
                data=form
            )
        
        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(
                    f"Collection {collection} not found",
                    status_code=status
                )
            raise GetTaskError(
                f"Post processing failed with status {status}",
                status_code=status
            )
        
        return response

    @require_api_key
    async def download_files(
        self,
        collection: str,
        file_type: str,
        page: Optional[int] = None,  # Changed from 0 to None
        page_size: Optional[int] = 100,
        folder: Optional[str] = None,
        url: Optional[bool] = True
    ) -> Dict[str, Any]:
        """
        Get list of signed urls to download files in collection, or download the files directly.

        Args:
            collection: Name of collection
            file_type: Type of files to download - must be either 'raw' or 'processed'
            page: Page number (optional). If None, downloads all pages. If specified, downloads only that page.
            page_size: Number of files to return per page (optional, defaults to 100)
            folder: If processed file type, which folder to download files from (optional)
            url: If True, return signed URLs; if False, download files directly (optional, defaults to True)

        Returns:
            API response as a dictionary containing list of download URLs (if url=True),
            or a dictionary with downloaded file information (if url=False)

        Raises:
            CollectionNotFoundError: If the collection is not found
            DownloadFilesError: If the API request fails due to unknown reasons
            ValueError: If file_type is not 'raw' or 'processed'
        """
        if file_type not in ['raw', 'processed']:
            raise ValueError(f"file_type must be either 'raw' or 'processed', got '{file_type}'")
        
        # Determine if we should download all pages or just one
        download_all_pages = (page is None)
        current_page = 0 if download_all_pages else page
        
        all_files = []
        
        while True:
            params = {"file_type": file_type, "page": current_page}
            
            if page_size is not None:
                params["page_size"] = page_size
            if folder is not None:
                params["folder"] = folder

            response, status = await self._client._terrakio_request("GET", f"collections/{collection}/download", params=params)

            if status != 200:
                if status == 404:
                    raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
                raise DownloadFilesError(f"Download files failed with status {status}", status_code=status)
            
            files = response.get('files', []) if isinstance(response, dict) else []
            
            if not files:
                break
            
            all_files.extend(files)
            
            # If user specified a specific page, only download that page
            if not download_all_pages:
                break
            
            current_page += 1
        
        if url:
            # Return URLs format matching original API response
            return {
                'collection': collection,
                'files': all_files,
                'total': len(all_files)
            }
        
        # Download all files
        downloaded_files = []
        
        async with aiohttp.ClientSession() as session:
            for file_info in all_files:
                try:
                    file_url = file_info.get('url')
                    filename = file_info.get('file', '')
                    group = file_info.get('group', '')
                    
                    if not file_url:
                        downloaded_files.append({
                            'filename': filename,
                            'group': group,
                            'error': 'No URL provided'
                        })
                        continue
                    
                    async with session.get(file_url) as file_response:
                        if file_response.status == 200:
                            content = await file_response.read()
                            
                            output_dir = folder if folder else "downloads"
                            if group:
                                output_dir = os.path.join(output_dir, group)
                            os.makedirs(output_dir, exist_ok=True)
                            filepath = os.path.join(output_dir, filename)
                            
                            with open(filepath, 'wb') as f:
                                f.write(content)
                            
                            downloaded_files.append({
                                'filename': filename,
                                'group': group,
                                'filepath': filepath,
                                'size': len(content)
                            })
                        else:
                            downloaded_files.append({
                                'filename': filename,
                                'group': group,
                                'error': f"Failed to download: HTTP {file_response.status}"
                            })
                except Exception as e:
                    downloaded_files.append({
                        'filename': file_info.get('file', 'unknown'),
                        'group': file_info.get('group', ''),
                        'error': str(e)
                    })
        
        return {
            'collection': collection,
            'downloaded_files': downloaded_files,
            'total': len(downloaded_files)
        }

    @require_api_key
    async def gen_and_process(
        self,
        collection: str,
        requests_file: Union[str, Any],
        output: str,
        folder: str,
        consumer: Union[str, Any],
        extra: Optional[Dict[str, Any]] = None,
        force_loc: Optional[bool] = False,
        skip_existing: Optional[bool] = True,
        server: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Generate data and run post-processing in a single task.

        Args:
            collection: Name of collection
            requests_file: Path to JSON file or file object containing request configurations
            output: Output type (str)
            folder: Folder to store output
            consumer: Path to post processing script or file object
            extra: Additional configuration parameters (optional)
            force_loc: Write data directly to the cloud under this folder (optional, defaults to False)
            skip_existing: Skip existing data (optional, defaults to True)
            server: Server to use (optional)

        Returns:
            API response as a dictionary containing task information

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails due to unknown reasons
        """
        await self.get_collection(collection = collection)

        upload_urls = await self._get_upload_url(collection=collection)
        url = upload_urls['url']
        
        # Handle requests_file - either file path (str) or file object
        if isinstance(requests_file, str):
            await self._upload_file(requests_file, url)
        else:
            # File object - read JSON and upload directly
            json_data = json.load(requests_file)
            await self._upload_json_data(json_data, url)

        # Handle consumer - either file path (str) or file object
        if isinstance(consumer, str):
            with open(consumer, 'rb') as f:
                consumer_content = f.read()
        else:
            # Assume it's a file object
            consumer_content = consumer.read()
        
        form = aiohttp.FormData()
        form.add_field('output', output)
        form.add_field('force_loc', str(force_loc).lower())
        form.add_field('skip_existing', str(skip_existing).lower())
        
        if server is not None:
            form.add_field('server', server)
        
        form.add_field('extra', json.dumps(extra or {}))
        form.add_field('folder', folder)
        form.add_field(
            'consumer',
            consumer_content,
            filename='consumer.py',
            content_type='text/x-python'
        )
        
        response, status = await self._client._terrakio_request(
            "POST",
            f"collections/{collection}/gen_and_process",
            data=form
        )
        
        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise GetTaskError(f"Gen and process failed with status {status}", status_code=status)

        return response




import asyncio
import concurrent.futures
import threading
import functools
import inspect
from typing import Optional, Dict, Any, Union, TYPE_CHECKING
from geopandas import GeoDataFrame
from shapely.geometry.base import BaseGeometry as ShapelyGeometry
from .async_client import AsyncClient

# Add type checking imports for better IDE support
if TYPE_CHECKING:
    from .endpoints.dataset_management import DatasetManagement
    from .endpoints.user_management import UserManagement
    from .endpoints.collections import Collections
    from .endpoints.group_management import GroupManagement
    from .endpoints.space_management import SpaceManagement
    from .endpoints.model_management import ModelManagement
    from .endpoints.auth import AuthClient


class SyncWrapper:
    """Generic synchronous wrapper with __dir__ support for runtime autocomplete."""
    
    def __init__(self, async_obj, sync_client):
        self._async_obj = async_obj
        self._sync_client = sync_client
    
    def __dir__(self):
        """Return list of attributes for autocomplete in interactive environments."""
        async_attrs = [attr for attr in dir(self._async_obj) if not attr.startswith('_')]
        wrapper_attrs = [attr for attr in object.__dir__(self) if not attr.startswith('_')]
        return list(set(async_attrs + wrapper_attrs))
    
    def __getattr__(self, name):
        """Dynamically wrap any method call to convert async to sync."""
        attr = getattr(self._async_obj, name)
        
        if callable(attr):
            @functools.wraps(attr)
            def sync_wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                if hasattr(result, '__await__'):
                    return self._sync_client._run_async(result)
                return result
            return sync_wrapper
        
        return attr


class SyncClient:
    """
    Thread-safe synchronous wrapper for AsyncClient.
    Uses a persistent event loop in a dedicated thread to avoid event loop conflicts.
    """

    # Add explicit type annotations for endpoint managers
    datasets: 'DatasetManagement'
    users: 'UserManagement' 
    collections: 'Collections'
    groups: 'GroupManagement'
    space: 'SpaceManagement'
    model: 'ModelManagement'
    auth: 'AuthClient'
    
    def __init__(self, url: Optional[str] = None, api_key: Optional[str] = None, verbose: bool = False):
        self._async_client = AsyncClient(url=url, api_key=api_key, verbose=verbose)
        self._context_entered = False
        self._closed = False
        
        # Thread and event loop management
        self._loop = None
        self._thread = None
        self._loop_ready = None
        self._loop_exception = None
        
        # Initialize endpoint managers with proper typing

        self.datasets = SyncWrapper(self._async_client.datasets, self)
        self.users = SyncWrapper(self._async_client.users, self)
        self.collections = SyncWrapper(self._async_client.collections, self)
        self.groups = SyncWrapper(self._async_client.groups, self)
        self.space = SyncWrapper(self._async_client.space, self)
        self.model = SyncWrapper(self._async_client.model, self)
        self.auth = SyncWrapper(self._async_client.auth, self)
        
        import atexit
        atexit.register(self._cleanup)
    
    def _ensure_event_loop(self) -> None:
        """Ensure we have a persistent event loop in a dedicated thread."""
        if self._loop is None or self._loop.is_closed():
            self._loop_ready = threading.Event()
            self._loop_exception = None
            
            def run_loop():
                """Run the event loop in a dedicated thread."""
                try:
                    # Create a new event loop for this thread
                    self._loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(self._loop)
                    
                    # Signal that the loop is ready
                    self._loop_ready.set()
                    
                    # Run the loop forever (until stopped)
                    self._loop.run_forever()
                except Exception as e:
                    self._loop_exception = e
                    self._loop_ready.set()
                finally:
                    # Clean up when the loop stops
                    if self._loop and not self._loop.is_closed():
                        self._loop.close()
            
            # Start the thread
            self._thread = threading.Thread(target=run_loop, daemon=True)
            self._thread.start()
            
            # Wait for the loop to be ready
            self._loop_ready.wait(timeout=10)
            
            if self._loop_exception:
                raise self._loop_exception
            
            if not self._loop_ready.is_set():
                raise RuntimeError("Event loop failed to start within timeout")
    
    def _run_async(self, coro):
        """
        Run async coroutine using persistent event loop.
        This is the core method that makes everything work.
        """
        # Ensure we have an event loop
        self._ensure_event_loop()
        
        if self._loop.is_closed():
            raise RuntimeError("Event loop is closed")
        
        # Create a future to get the result back from the event loop thread
        future = concurrent.futures.Future()
        
        async def run_with_context():
            """Run the coroutine with proper context management."""
            try:
                # Ensure the async client is properly initialized
                await self._ensure_context()
                
                # Run the actual coroutine
                result = await coro
                
                # Set the result on the future
                future.set_result(result)
            except Exception as e:
                # Set the exception on the future
                future.set_exception(e)
        
        # Schedule the coroutine on the persistent event loop
        self._loop.call_soon_threadsafe(
            lambda: asyncio.create_task(run_with_context())
        )

        return future.result()
    
    async def _ensure_context(self) -> None:
        """Ensure the async client context is entered."""
        if not self._context_entered and not self._closed:
            await self._async_client.__aenter__()
            self._context_entered = True
    
    async def _exit_context(self) -> None:
        """Exit the async client context."""
        if self._context_entered and not self._closed:
            await self._async_client.__aexit__(None, None, None)
            self._context_entered = False
    
    def close(self) -> None:
        """Close the underlying async client session and stop the event loop."""
        if not self._closed:
            if self._loop and not self._loop.is_closed():
                # Schedule cleanup on the event loop
                future = concurrent.futures.Future()
                
                async def cleanup():
                    """Clean up the async client."""
                    try:
                        await self._exit_context()
                        future.set_result(None)
                    except Exception as e:
                        future.set_exception(e)
                
                # Run cleanup
                self._loop.call_soon_threadsafe(
                    lambda: asyncio.create_task(cleanup())
                )
                
                # Wait for cleanup to complete
                try:
                    future.result(timeout=10)
                except:
                    pass  # Ignore cleanup errors
                
                # Stop the event loop
                self._loop.call_soon_threadsafe(self._loop.stop)
                
                # Wait for thread to finish
                if self._thread and self._thread.is_alive():
                    self._thread.join(timeout=5)
            
            self._closed = True
    
    def _cleanup(self) -> None:
        """Internal cleanup method called by atexit."""
        if not self._closed:
            try:
                self.close()
            except Exception:
                pass  # Ignore cleanup errors
    
    def __dir__(self):
        """Return list of attributes for autocomplete in interactive environments."""
        default_attrs = [attr for attr in object.__dir__(self) if not attr.startswith('_')]
        async_client_attrs = [attr for attr in dir(self._async_client) if not attr.startswith('_')]
        endpoint_attrs = ['datasets', 'users', 'collections', 'groups', 'space', 'model', 'auth']
        all_attrs = default_attrs + async_client_attrs + endpoint_attrs
        return list(set(all_attrs))
    
    # Your existing methods with proper type annotations
    def geoquery(
        self,
        expr: str,
        feature: Union[Dict[str, Any], ShapelyGeometry],
        in_crs: str = "epsg:4326",
        out_crs: str = "epsg:4326",
        resolution: int = -1,
        geom_fix: bool = False,
        **kwargs
    ):
        """Compute WCS query for a single geometry (synchronous version)."""
        coro = self._async_client.geoquery(
            expr=expr,
            feature=feature,
            in_crs=in_crs,
            out_crs=out_crs,
            output="netcdf",
            resolution=resolution,
            geom_fix=geom_fix,
            **kwargs
        )
        return self._run_async(coro)

    def zonal_stats(self, *args, **kwargs) -> GeoDataFrame:
        """Proxy to async zonal_stats with full argument passthrough (sync wrapper)."""
        coro = self._async_client.zonal_stats(*args, **kwargs)
        return self._run_async(coro)
    
    def create_dataset_file(self, *args, **kwargs) -> dict:
        """Proxy to async create_dataset_file with full argument passthrough (sync wrapper)."""
        coro = self._async_client.create_dataset_file(*args, **kwargs)
        return self._run_async(coro)


    def geo_queries(self, *args, **kwargs) -> Union[float, GeoDataFrame]:
        """Proxy to async geo_queries with full argument passthrough (sync wrapper)."""
        coro = self._async_client.geo_queries(*args, **kwargs)
        return self._run_async(coro)
    
    # Context manager support
    def __enter__(self) -> 'SyncClient':
        """Context manager entry."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Context manager exit."""
        self.close()
    
    def __del__(self) -> None:
        """Destructor to ensure session is closed."""
        if not self._closed:
            try:
                self._cleanup()
            except Exception:
                pass


import time
from typing import Any, Dict, List, Optional

from dateutil import parser
from rich.progress import (
    BarColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)

from ...exceptions import (
    CancelAllTasksError,
    CancelCollectionTasksError,
    CancelTaskError,
    CollectionNotFoundError,
    GetTaskError,
    ListTasksError,
    TaskNotFoundError,
)
from ...helper.decorators import require_api_key


class TasksMixin:
    """Task management operations."""
    
    async def track_progress(self, task_id):
        task_info = await self.get_task(task_id=task_id)
        number_of_jobs = task_info["task"]["total"]
        start_time = parser.parse(task_info["task"]["createdAt"])
        
        self.console.print(f"[bold cyan]Tracking task: {task_id}[/bold cyan]")
        
        completed_jobs_info = []
        
        def get_job_description(job_info, include_status=False):
            if not job_info:
                return "No job info"
            
            service = job_info.get("service", "Unknown service")
            desc = service
            
            if include_status:
                status = job_info.get("status", "unknown")
                desc += f" - {status}"
            
            return desc
        
        progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
        )
        
        with progress:
            last_completed_count = 0
            current_job_task = None
            current_job_description = None
            
            while len(completed_jobs_info) < number_of_jobs:
                task_info = await self.get_task(task_id=task_id)
                completed_number = task_info["task"]["completed"]
                current_job_info = task_info["currentJob"]
                
                # Check task-level status first
                task_status = task_info.get("task", {}).get("status", "").lower()
                if task_status in ["error", "failed", "cancelled"]:
                    progress.stop()
                    error_msg = f"\n[bold red]Task {task_status}![/bold red]"
                    
                    # Try to get error details from currentJob
                    if current_job_info and current_job_info.get("error"):
                        error_msg += f"\n[red]Reason: {current_job_info['error']}[/red]"
                    
                    self.console.print(error_msg)
                    return None
                
                # Check for error/cancelled/failed status in current job
                if current_job_info:
                    status = current_job_info.get("status", "").lower()
                    if status in ["error", "failed", "cancelled"]:
                        progress.stop()
                        error_msg = f"\n[bold red]Job {status}![/bold red]"
                        
                        # Get error details if available
                        if current_job_info.get("error"):
                            error_msg += f"\n[red]Reason: {current_job_info['error']}[/red]"
                        
                        self.console.print(error_msg)
                        return None
                
                if completed_number > last_completed_count:
                    if current_job_task is not None:
                        completed_description = current_job_description.replace(" - pending", "").replace(" - running", "").replace(" - waiting", "")
                        completed_description += " - completed"
                        
                        progress.update(
                            current_job_task,
                            description=f"[{last_completed_count + 1}/{number_of_jobs}] {completed_description}",
                            completed=100
                        )
                        completed_jobs_info.append({
                            "task": current_job_task,
                            "description": completed_description,
                            "job_number": last_completed_count + 1
                        })
                        current_job_task = None
                        current_job_description = None
                    
                    last_completed_count = completed_number
                
                if current_job_info:
                    status = current_job_info["status"]
                    current_job_description = get_job_description(current_job_info, include_status=True)
                    
                    total_value = current_job_info.get("total", 0)
                    completed_value = current_job_info.get("completed", 0)
                    
                    if total_value == -9999:
                        percent = 0
                    elif total_value > 0:
                        percent = int(completed_value / total_value * 100)
                    else:
                        percent = 0
                    
                    if current_job_task is None:
                        current_job_task = progress.add_task(
                            f"[{completed_number + 1}/{number_of_jobs}] {current_job_description}",
                            total=100,
                            start_time=start_time
                        )
                    else:
                        progress.update(
                            current_job_task,
                            description=f"[{completed_number + 1}/{number_of_jobs}] {current_job_description}",
                            completed=percent
                        )
                    
                    if status == "Completed":
                        completed_description = current_job_description.replace(" - pending", "").replace(" - running", "").replace(" - waiting", "")
                        completed_description += " - completed"
                        progress.update(
                            current_job_task, 
                            description=f"[{completed_number + 1}/{number_of_jobs}] {completed_description}",
                            completed=100
                        )
                
                if completed_number == number_of_jobs and current_job_info is None:
                    if current_job_task is not None:
                        completed_description = current_job_description.replace(" - pending", "").replace(" - running", "").replace(" - waiting", "")
                        completed_description += " - completed"
                        progress.update(
                            current_job_task,
                            description=f"[{number_of_jobs}/{number_of_jobs}] {completed_description}",
                            completed=100
                        )
                        completed_jobs_info.append({
                            "task": current_job_task,
                            "description": completed_description,
                            "job_number": number_of_jobs
                        })
                    break
                
                time.sleep(10)
        
        self.console.print(f"[bold green]All {number_of_jobs} jobs finished![/bold green]")

    @require_api_key
    async def list_tasks(
        self,
        limit: Optional[int] = 10,
        page: Optional[int] = 0
    ) -> List[Dict[str, Any]]:
        """
        List tasks for the current user.

        Args:
            limit: Number of tasks to return (optional, defaults to 10)
            page: Page number (optional, defaults to 0)
        
        Returns:
            API response as a list of dictionaries containing task information
            
        Raises:
            ListTasksError: If the API request fails due to unknown reasons
        """
        params = {
            "limit": limit,
            "page": page
        }
        response, status = await self._client._terrakio_request("GET", "tasks", params=params)

        if status != 200:
            raise ListTasksError(f"List tasks failed with status {status}", status_code=status)

        return response
        
    @require_api_key
    async def get_task(
        self,
        task_id: str
    ) -> Dict[str, Any]:
        """
        Get task information by task ID.

        Args:
            task_id: ID of task to track
        
        Returns:
            API response as a dictionary containing task information

        Raises:
            TaskNotFoundError: If the task is not found
            GetTaskError: If the API request fails due to unknown reasons
        """
        response, status = await self._client._terrakio_request("GET", f"tasks/info/{task_id}")

        if status != 200:
            if status == 404:
                raise TaskNotFoundError(f"Task {task_id} not found", status_code=status)
            raise GetTaskError(f"Get task failed with status {status}", status_code=status)
        
        return response

    @require_api_key
    async def cancel_task(
        self,
        task_id: str
    ) -> Dict[str, Any]:
        """
        Cancel a task by task ID.

        Args:
            task_id: ID of task to cancel

        Returns:
            API response as a dictionary containing task information

        Raises:
            TaskNotFoundError: If the task is not found
            CancelTaskError: If the API request fails due to unknown reasons
        """
        response, status = await self._client._terrakio_request("POST", f"tasks/cancel/{task_id}")
        
        if status != 200:
            if status == 404:
                raise TaskNotFoundError(f"Task {task_id} not found", status_code=status)
            raise CancelTaskError(f"Cancel task failed with status {status}", status_code=status)

        return response

    @require_api_key
    async def cancel_collection_tasks(
        self,
        collection: str
    ) -> Dict[str, Any]:
        """
        Cancel all tasks for a collection.

        Args:
            collection: Name of collection

        Returns:
            API response as a dictionary containing task information for the collection

        Raises:
            CollectionNotFoundError: If the collection is not found
            CancelCollectionTasksError: If the API request fails due to unknown reasons
        """
        response, status = await self._client._terrakio_request("POST", f"collections/{collection}/cancel")
        
        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise CancelCollectionTasksError(f"Cancel collection tasks failed with status {status}", status_code=status)
    
        return response

    @require_api_key
    async def cancel_all_tasks(
        self
    ) -> Dict[str, Any]:
        """
        Cancel all tasks for the current user.

        Returns:
            API response as a dictionary containing task information for all tasks

        Raises:
            CancelAllTasksError: If the API request fails due to unknown reasons
        """
        response, status = await self._client._terrakio_request("POST", "tasks/cancel")

        if status != 200:
            raise CancelAllTasksError(f"Cancel all tasks failed with status {status}", status_code=status)

        return response




import os
from typing import Any, Dict, Optional

from ...exceptions import (
    CollectionNotFoundError,
    GetTaskError,
)
from ...helper.decorators import require_api_key


class ZonalStatsMixin:
    """Zonal statistics operations."""
    
    @require_api_key
    async def zonal_stats(
        self,
        collection: str,
        id_property: str,
        column_name: str,
        expr: str,
        resolution: Optional[int] = 1,
        in_crs: Optional[str] = "epsg:4326",
        out_crs: Optional[str] = "epsg:4326"
    ) -> Dict[str, Any]:
        """
        Run zonal stats over uploaded geojson collection.

        Args:
            collection: Name of collection
            id_property: Property key in geojson to use as id
            column_name: Name of new column to add
            expr: Terrak.io expression to evaluate
            resolution: Resolution of request (optional, defaults to 1)
            in_crs: CRS of geojson (optional, defaults to "epsg:4326")
            out_crs: Desired output CRS (optional, defaults to "epsg:4326")

        Returns:
            API response as a dictionary containing task information

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails due to unknown reasons
        """
        payload = {
            "id_property": id_property,
            "column_name": column_name,
            "expr": expr,
            "resolution": resolution,
            "in_crs": in_crs,
            "out_crs": out_crs
        }
        
        response, status = await self._client._terrakio_request("POST", f"collections/{collection}/zonal_stats", json=payload)

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise GetTaskError(f"Zonal stats failed with status {status}", status_code=status)
        
        return response

    @require_api_key
    async def zonal_stats_transform(
        self,
        collection: str,
        consumer: str
    ) -> Dict[str, Any]:
        """
        Transform raw data in collection. Creates a new collection.

        Args:
            collection: Name of collection
            consumer: Post processing script (file path or script content)

        Returns:
            API response as a dictionary containing task information

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails due to unknown reasons
        """
        if os.path.isfile(consumer):
            with open(consumer, 'r') as f:
                script_content = f.read()
        else:
            script_content = consumer

        files = {
            'consumer': ('script.py', script_content, 'text/plain')
        }
        
        response, status = await self._client._terrakio_request(
            "POST", 
            f"collections/{collection}/transform", 
            files=files
        )

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise GetTaskError(f"Transform failed with status {status}", status_code=status)
        
        return response




from rich.console import Console

from .collections import CollectionsMixin
from .common import Dataset_Dtype, OutputTypes, Region, regions
from .data_operations import DataOperationsMixin
from .generation import DataGenerationMixin
from .ingestion import IngestionMixin
from .tasks import TasksMixin
from .zonal_stats import ZonalStatsMixin

# Import methods directly for better IDE navigation support
from .collections import (
    CollectionsMixin as _CollectionsMixin,
)
from .tasks import (
    TasksMixin as _TasksMixin,
)
from .data_operations import (
    DataOperationsMixin as _DataOperationsMixin,
)
from .zonal_stats import (
    ZonalStatsMixin as _ZonalStatsMixin,
)
from .ingestion import (
    IngestionMixin as _IngestionMixin,
)
from .generation import (
    DataGenerationMixin as _DataGenerationMixin,
)


class Collections(
    CollectionsMixin,
    TasksMixin,
    DataOperationsMixin,
    ZonalStatsMixin,
    DataGenerationMixin,
    IngestionMixin
):
    """
    Collections and collection management client.
    
    This class provides methods for managing collections, tasks, data operations,
    zonal statistics, data generation, and ingestion operations.
    
    All methods are inherited from the following mixins:
    - CollectionsMixin: Collection CRUD operations (create_collection, get_collection, list_collections, delete_collection)
    - TasksMixin: Task management operations (track_progress, list_tasks, get_task, cancel_task, etc.)
    - DataOperationsMixin: Data generation and processing (generate_data, post_processing, gen_and_process, download_files, upload_artifacts)
    - ZonalStatsMixin: Zonal statistics operations (zonal_stats, zonal_stats_transform)
    - DataGenerationMixin: Data generation operations (training_samples, dataset, tiles, polygons)
    - IngestionMixin: Data ingestion and visualization (create_pyramids, tif)
    
    Note: Methods are defined in their respective mixin modules:
    - Collection methods: .collections module
    - Task methods: .tasks module
    - Data operation methods: .data_operations module
    - Zonal stats methods: .zonal_stats module
    - Data generation methods: .generation module
    - Ingestion methods: .ingestion module
    """
    
    def __init__(self, client):
        self._client = client
        self.console = Console()
        self.OutputTypes = OutputTypes
        self.Region = Region
        self.regions = regions
        self.Dataset_Dtype = Dataset_Dtype

# Explicitly re-export for better IDE support
__all__ = ['Collections', 'OutputTypes', 'Region', 'Dataset_Dtype', 'regions']



import enum
import json
import dateutil.parser
import geopandas as gpd
import shapely.geometry
import typer
from typing import Any

class OutputTypes(enum.Enum):
    geotiff = 'geotiff'
    png = 'png'
    netcdf = 'netcdf'
    json = 'json'
    json_v2 = 'json_v2'
    csv = 'csv'

class Region(str, enum.Enum):
    aus = "aus"
    eu = "eu"
    us = "us"

regions = {
    Region.aus : {
        "name" : "australia-southeast1", 
        "url" : "https://terrakio-server-candidate-573248941006.australia-southeast1.run.app", 
        "bucket" : "terrakio-mass-requests"
    },
    
    Region.eu : {
        "name" : "europe-west4", 
        "url" : "https://terrakio-server-candidate-573248941006.europe-west4.run.app", 
        "bucket" : "terrakio-mass-requests-eu"
    },
    
    Region.us : {
        "name" : "us-central1", 
        "url" : "https://terrakio-server-candidate-573248941006.us-central1.run.app", 
        "bucket" : "terrakio-mass-requests-us"
    },

}

class Dataset_Dtype(enum.Enum):
    uint8 = 'uint8'
    float32 = 'float32'

def tile_generator(x_min, y_min, x_max, y_max, aoi, crs, res, tile_size, expression, output, fully_cover=True):
    i_max = int((x_max-x_min)/(tile_size*res))
    j_max = int((y_max-y_min)/(tile_size*res))
    if fully_cover:
        i_max += 1
        j_max += 1
    for j in range(0, int(j_max)):
        for i in range(0, int(i_max)):
            x = x_min + i*(tile_size*res)
            y = y_max - j*(tile_size*res)
            bbox = shapely.geometry.box(x, y-(tile_size*res), x + (tile_size*res), y)
            if not aoi.geometry[0].intersects(bbox):
                continue
            feat  = {"type": "Feature", "geometry": bbox.__geo_interface__}
            data = {
                "feature": feat,
                "in_crs": crs,
                "out_crs": crs,
                "resolution": res,
                "expr" : expression,
                "output" : output.value,
            }
            yield data, i , j


def get_bounds(aoi, crs, to_crs = None):
    gdf : gpd.GeoDataFrame = gpd.read_file(aoi)
    gdf = gdf.set_crs(crs, allow_override=True)
    if to_crs:
        gdf = gdf.to_crs(to_crs)
    bounds = gdf.geometry[0].bounds
    return *bounds, gdf

def validate_date(date: str) -> str:
    try:
        date = dateutil.parser.parse(date)
        return date
    except ValueError:
        print(f"Invalid date: {date}")
        raise typer.BadParameter(f"Invalid date: {date}")

def make_json_serializable(obj):
    """Convert non-JSON-serializable types to JSON-serializable equivalents."""
    if isinstance(obj, dict):
        return {k: make_json_serializable(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [make_json_serializable(item) for item in obj]
    elif hasattr(obj, 'isoformat'):  # Timestamp, datetime, date, time
        return obj.isoformat()
    elif hasattr(obj, 'item'):  # numpy types
        return obj.item()
    else:
        return obj




import math
import os
import subprocess
import tempfile
from typing import Any, Dict, List, Optional

import numpy as np
import pyproj
import rasterio as rio
import snappy
import typer
from dateutil import parser
from rasterio.windows import from_bounds
from rich.progress import (
    BarColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
)

from ...exceptions import (
    CollectionAlreadyExistsError,
    CollectionNotFoundError,
    GetTaskError,
)
from ...helper.decorators import require_api_key


class IngestionMixin:
    """Data ingestion and visualization operations."""
    
    @require_api_key
    async def create_pyramids(
        self,
        name: str,
        levels: int,
        config: Dict[str, Any],
        collection: Optional[str] = None
    ) -> Dict[str, Any]:
        """
        Create pyramid tiles for a dataset.

        Args:
            name: Dataset name
            levels: Maximum zoom level for pyramid (e.g., 8)
            config: Dictionary of configuration key-value pairs
            collection: Name of collection (optional, defaults to "{name}-pyramids")

        Returns:
            API response with task_id

        Raises:
            GetTaskError: If the API request fails
            CollectionNotFoundError: If the collection is not found
            CreateCollectionError: If collection creation fails
        """
        if collection is None or collection == "":
            collection = f"{name}-pyramids"
            try:
                await self.create_collection(collection=collection)
            except CollectionAlreadyExistsError:
                # Collection already exists, continue with it
                pass
        else:
            try:
                await self.get_collection(collection=collection)
            except CollectionNotFoundError:
                await self.create_collection(collection=collection)

        pyramid_request = {
            'collection_name': collection,
            'name': name,
            'max_zoom': levels,
            **config
        }

        response, status = await self._client._terrakio_request(
            "POST",
            "tasks/pyramids",
            json=pyramid_request
        )

        if status != 200:
            raise GetTaskError(
                f"Pyramid creation failed with status {status}: {response}", 
                status_code=status
            )
        
        task_id = response["task_id"]
        await self.track_progress(task_id)

        return {"task_id": task_id}

    @require_api_key
    async def tif(
        self,
        file: str,
        dataset: str,
        product: List[str],
        bucket: str,
        path: str,
        no_data: float,
        max_zoom: int,
        date: str,
        add_config: bool = True,
        generate_pyramids: bool = True,
        geot: Optional[List[float]] = None,
        no_interactive: bool = False,
        tile_size: int = 400,
        update_config: bool = True,
    ) -> Dict[str, Any]:
        """
        Ingest a tif file and optionally generate pyramids.

        Args:
            file: Path to the tif file to ingest
            dataset: Dataset name
            product: List of product names
            bucket: Storage bucket
            path: Storage path
            no_data: No data value
            max_zoom: Maximum zoom level
            date: Date string
            add_config: Add dataset configuration (default: True)
            generate_pyramids: Generate pyramids after ingestion (default: True)
            geot: Geotransform parameters (optional)
            no_interactive: Non-interactive mode (default: False)
            tile_size: Size of tiles (default: 400)
            update_config: Update config with new date (default: True)

        Returns:
            API response with task_id

        Raises:
            GetTaskError: If the API request fails
        """
        with rio.open(file) as src:
            if len(product) != src.meta["count"]:
                print("[bold red]Products don't match number of bands[/bold red]")
                raise typer.Exit(code=2)
            meta = src.meta
            dtype = meta["dtype"]
            transform = meta["transform"]
            if not geot:
                geot = [transform[2], transform[0], transform[1], transform[5], transform[3], transform[4]]
            x0, y0 = geot[0], geot[3]
            proj = pyproj.Proj(meta["crs"])

            j_max = int(math.ceil(meta['height'] / tile_size))
            i_max = int(math.ceil(meta['width'] / tile_size))

            if add_config:
                await self._client.datasets.create_dataset(
                    name=dataset,
                    products=product,
                    dates_iso8601=[date],
                    bucket=bucket,
                    path=f"{path}/%s_%s_%03d_%03d_%02d.snp",
                    data_type=dtype,
                    no_data=no_data,
                    i_max=i_max,
                    j_max=j_max,
                    y_size=tile_size,
                    x_size=tile_size,
                    proj4=proj.definition_string(),
                    abstract="",
                    geotransform=geot,
                    max_zoom=0,
                )
            else:
                if not no_interactive and update_config:
                    await self._client.datasets.update_dataset(
                        name=dataset,
                        append=True,
                        dates_iso8601=[date]
                    )
    
            with tempfile.TemporaryDirectory() as tmpdirname:
                progress = Progress(
                    TextColumn("[progress.description]{task.description}"),
                    BarColumn(),
                    TaskProgressColumn(),
                    TimeElapsedColumn(),
                )
                
                with progress:
                    task = progress.add_task(description="[bold green]Writing tiles...[/bold green]", total=j_max*i_max)
                    for j in range(0, j_max+1):
                        for i in range(0, i_max+1):
                            left = x0 + i*tile_size*geot[1]
                            right = x0 + (i+1)*tile_size*geot[1]
                            top = y0 - j*tile_size*geot[1]
                            bottom = y0 - (j+1)*tile_size*geot[1]

                            data = src.read(window=from_bounds(left, bottom, right, top, src.transform), boundless=True)
                            for band in range(src.meta["count"]):
                                out = data[band]
                                if np.all(np.isnan(out)):
                                    progress.update(task, advance=1)
                                    continue
                            
                                out[np.isnan(out)] = no_data
                                out_bytes = out.tobytes()
                                
                                date_obj = parser.parse(date)
                                with open(f"{tmpdirname}/{product[band]}_{date_obj.strftime('%Y%m%d%H%M%S')}_{i:03d}_{j:03d}_00.snp", 'wb') as f:
                                    comp = snappy.compress(out_bytes)
                                    f.write(comp)
                                progress.update(task, advance=1)

                subprocess.run(
                    ["gsutil", "-m", "cp", "-r", "*", f"gs://{bucket}/{path}"],
                    cwd=f"{tmpdirname}",
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )

        if generate_pyramids:
            await self.create_pyramids(
                name = dataset,
                levels = max_zoom,
                config = {
                    "products": product,
                    "dates_iso8601": [date],
                }
            )




class APIError(Exception):
    """Exception raised for errors in the API responses."""
    
    def __init__(self, message, status_code=None):
        super().__init__(message)
        self.status_code = status_code

class ConfigurationError(Exception):
    """Exception raised for errors in the configuration."""
    pass


class DownloadError(Exception):
    """Exception raised for errors during data download."""
    pass


class ValidationError(Exception):
    """Exception raised for invalid request parameters."""
    pass

class NetworkError(Exception):
    """Exception raised for network errors."""
    pass

class AuthenticationExpireError(Exception):
   """Exception raised for authentication expire errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class QuotaError(Exception):
   """Exception raised for quota errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class UserInfoError(Exception):
   """Exception raised for user info errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class APIKeyError(Exception):
   """Exception raised for API key errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class RefreshAPIKeyError(Exception):
   """Exception raised for refresh API key errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class InvalidUsernamePasswordError(Exception):
   """Exception raised for invalid username or password errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class LoginError(Exception):
   """Exception raised for login errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ResetPasswordError(Exception):
   """Exception raised for reset password errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class SignupError(Exception):
   """Exception raised for signup errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class InvalidEmailFormatError(Exception):
   """Exception raised for invalid email format errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class EmailAlreadyExistsError(Exception):
   """Exception raised for email already exists errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class UserNotFoundError(Exception):
   """Exception raised for user not found errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetUserByIdError(Exception):
   """Exception raised for get user by id errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetUserByEmailError(Exception):
   """Exception raised for get user by email errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ListUsersError(Exception):
   """Exception raised for list users errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class EditUserError(Exception):
   """Exception raised for edit user errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ResetQuotaError(Exception):
   """Exception raised for reset quota errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DeleteUserError(Exception):
   """Exception raised for delete user errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ChangeRoleError(Exception):
   """Exception raised for change role errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ListGroupsError(Exception):
   """Exception raised for list groups errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetGroupError(Exception):
   """Exception raised for get group errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetGroupDatasetsError(Exception):
   """Exception raised for get group datasets errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class NoDatasetsFoundForGroupError(Exception):
   """Exception raised for no datasets found for group errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CreateGroupError(Exception):
   """Exception raised for create group errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DeleteGroupError(Exception):
   """Exception raised for delete group errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class AddUserToGroupError(Exception):
   """Exception raised for add user to group errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class AddGroupToDatasetError(Exception):
   """Exception raised for add group to dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class AddUserToDatasetError(Exception):
   """Exception raised for add user to dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class RemoveUserFromGroupError(Exception):
   """Exception raised for remove user from group errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class RemoveUserFromDatasetError(Exception):
   """Exception raised for remove user from dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GroupNotFoundError(Exception):
   """Exception raised for group not found errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GroupPermissionError(Exception):
   """Exception raised for group permission errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CommandPermissionError(Exception):
   """Exception raised for command permission errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DatasetNotFoundError(Exception):
   """Exception raised for dataset not found errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class RemoveGroupFromDatasetError(Exception):
   """Exception raised for remove group from dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ListDatasetsError(Exception):
   """Exception raised for list datasets errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetUsersByRoleError(Exception):
   """Exception raised for get users by role errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class RoleDoNotExistError(Exception):
   """Exception raised for role does not exist errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ChangeRoleError(Exception):
   """Exception raised for change role errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetDatasetError(Exception):
   """Exception raised for get dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DatasetPermissionError(Exception):
   """Exception raised for dataset permission errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CreateDatasetError(Exception):
   """Exception raised for create dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DatasetAlreadyExistsError(Exception):
   """Exception raised for dataset already exists errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DeleteDatasetError(Exception):
   """Exception raised for delete dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class OverwriteDatasetError(Exception):
   """Exception raised for overwrite dataset errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ListCollectionsError(Exception):
   """Exception raised for list collections errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetCollectionError(Exception):
   """Exception raised for get collection errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CreateCollectionError(Exception):
   """Exception raised for create collection errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CollectionNotFoundError(Exception):
   """Exception raised for collection not found errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CollectionAlreadyExistsError(Exception):
   """Exception raised for collection already exists errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class InvalidCollectionTypeError(Exception):
   """Exception raised for invalid collection type errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DeleteCollectionError(Exception):
   """Exception raised for delete collection errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class ListTasksError(Exception):
   """Exception raised for list tasks errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class UploadRequestsError(Exception):
   """Exception raised for upload requests errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class UploadArtifactsError(Exception):
   """Exception raised for upload artifacts errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GetTaskError(Exception):
   """Exception raised for get task errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class TaskNotFoundError(Exception):
   """Exception raised for task not found errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class DownloadFilesError(Exception):
   """Exception raised for download files errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CancelTaskError(Exception):
   """Exception raised for cancel task errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CancelCollectionTasksError(Exception):
   """Exception raised for cancel collection tasks errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class CancelAllTasksError(Exception):
   """Exception raised for cancel all tasks errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code

class GeoQueryError(Exception):
   """Exception raised for geo query errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code
    
class QuotaInsufficientError(Exception):
   """Exception raised for quota insufficient errors."""
   def __init__(self, message: str, status_code: int = None):
       super().__init__(message)
       self.status_code = status_code


from typing import Any, Dict, List, Optional

from ...exceptions import (
    CollectionAlreadyExistsError,
    CollectionNotFoundError,
    CreateCollectionError,
    DeleteCollectionError,
    GetCollectionError,
    InvalidCollectionTypeError,
    ListCollectionsError,
)
from ...helper.decorators import require_api_key


class CollectionsMixin:
    """Collection CRUD operations."""
    
    @require_api_key
    async def create_collection(
        self, 
        collection: str, 
        bucket: Optional[str] = None, 
        location: Optional[str] = None, 
        collection_type: str = "basic"
    ) -> Dict[str, Any]:
        """
        Create a collection for the current user.

        Args:
            collection: The name of the collection (required)
            bucket: The bucket to use (optional, admin only)
            location: The location to use (optional, admin only)
            collection_type: The type of collection to create (optional, defaults to "basic")
            
        Returns:
            API response as a dictionary containing the collection id
            
        Raises:
            CollectionAlreadyExistsError: If the collection already exists
            InvalidCollectionTypeError: If the collection type is invalid
            CreateCollectionError: If the API request fails due to unknown reasons
        """
        payload = {
            "collection_type": collection_type
        }
        
        if bucket is not None:
            payload["bucket"] = bucket
        
        if location is not None:
            payload["location"] = location
        
        response, status = await self._client._terrakio_request("POST", f"collections/{collection}", json=payload)
        if status != 200:
            if status == 400 or status == 409:
                raise CollectionAlreadyExistsError(f"Collection {collection} already exists", status_code=status)
            if status == 422:
                raise InvalidCollectionTypeError(f"Invalid collection type: {collection_type}", status_code=status)
            raise CreateCollectionError(f"Create collection failed with status {status}", status_code=status)

        return response

    @require_api_key
    async def get_collection(self, collection: str) -> Dict[str, Any]:
        """
        Get a collection by name.

        Args:
            collection: The name of the collection to retrieve(required)
            
        Returns:
            API response as a dictionary containing collection information
            
        Raises:
            CollectionNotFoundError: If the collection is not found
            GetCollectionError: If the API request fails due to unknown reasons
        """
        response, status = await self._client._terrakio_request("GET", f"collections/{collection}")

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise GetCollectionError(f"Get collection failed with status {status}", status_code=status)
        
        return response

    @require_api_key
    async def list_collections(
        self,
        collection_type: Optional[str] = None,
        limit: Optional[int] = 10,
        page: Optional[int] = 0
    ) -> List[Dict[str, Any]]:
        """
        List collections for the current user.

        Args:
            collection_type: Filter by collection type (optional)
            limit: Number of collections to return (optional, defaults to 10)
            page: Page number (optional, defaults to 0)
            
        Returns:
            API response as a list of dictionaries containing collection information
            
        Raises:
            ListCollectionsError: If the API request fails due to unknown reasons
        """
        params = {}
        
        if collection_type is not None:
            params["collection_type"] = collection_type
        
        if limit is not None:
            params["limit"] = limit
            
        if page is not None:
            params["page"] = page
        
        response, status = await self._client._terrakio_request("GET", "collections", params=params)

        if status != 200:
            raise ListCollectionsError(f"List collections failed with status {status}", status_code=status)
        
        return response

    @require_api_key
    async def delete_collection(
        self, 
        collection: str, 
        full: Optional[bool] = False, 
        outputs: Optional[list] = [], 
        data: Optional[bool] = False
    ) -> Dict[str, Any]:
        """
        Delete a collection by name.

        Args:
            collection: The name of the collection to delete (required)
            full: Delete the full collection (optional, defaults to False)
            outputs: Specific output folders to delete (optional, defaults to empty list)
            data: Whether to delete raw data (xdata folder) (optional, defaults to False)
            
        Returns:
            API response as a dictionary confirming deletion
            
        Raises:
            CollectionNotFoundError: If the collection is not found
            DeleteCollectionError: If the API request fails due to unknown reasons
        """
        params = {
            "full": str(full).lower(),
            "data": str(data).lower()
        }
        
        if outputs:
            params["outputs"] = outputs
        
        response, status = await self._client._terrakio_request("DELETE", f"collections/{collection}", params=params)

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise DeleteCollectionError(f"Delete collection failed with status {status}", status_code=status)

        return response




import json
from typing import Any, Dict, List, Optional

import geopandas as gpd
import pyproj

from ...exceptions import (
    CollectionNotFoundError,
    GetTaskError,
    QuotaInsufficientError,
)
from ...helper.decorators import require_api_key
from .common import (
    Dataset_Dtype,
    OutputTypes,
    get_bounds as _get_bounds,
    make_json_serializable as _make_json_serializable,
    tile_generator as _tile_generator,
    validate_date as _validate_date,
)


class DataGenerationMixin:
    """Data generation operations for collections."""

    @require_api_key
    async def training_samples(
        self,
        collection: str,
        aoi: str,
        expression_x: str,
        filter_x: str = "skip",
        filter_x_rate: float = 1,
        expression_y: str = "skip",
        filter_y: str = "skip",
        filter_y_rate: float = 1,
        samples: int = 1000,
        tile_size: float = 256,
        crs: str = "epsg:3577",
        res: float = 10,
        res_y: float = None,
        skip_test: bool = False,
        time_range: tuple[str, str] = None,
        time_precision: str = None,
        server: str = None,
        extra_filters: list[str] = None,
        extra_filters_rate: list[float] = None,
        extra_filters_res: list[float] = None,
        try_limit: int = 1000,
        allow_overlap: bool = True
    ) -> dict:
        """
        Generate an AI dataset using specified parameters.

        Args:
            collection: The collection name where we save the results
            aoi: Path to GeoJSON file containing area of interest
            expression_x: Expression for X data (features)
            filter_x: Filter expression for X data (default: "skip")
            filter_x_rate: Filter rate for X data (default: 1)
            expression_y: Expression for Y data (labels) (default: "skip")
            filter_y: Filter expression for Y data (default: "skip")
            filter_y_rate: Filter rate for Y data (default: 1)
            samples: Number of samples to generate (default: 1000)
            tile_size: Size of tiles in pixels (default: 256)
            crs: Coordinate reference system (default: "epsg:3577")
            res: Resolution for X data (default: 10)
            res_y: Resolution for Y data, defaults to res if None
            skip_test: Skip expression validation test (default: False)
            time_range: Tuple of (start_datetime, end_datetime) in ISO format (e.g., ("2023-01-01T00:00:00Z", "2025-01-01T00:00:00Z"))
            time_precision: Precision for random time sampling: 'year', 'month', 'day', 'hour', 'minute', or 'second'
            server: Server to use for processing
            extra_filters: Additional filter expressions
            extra_filters_rate: Rates for additional filters
            extra_filters_res: Resolutions for additional filters
            try_limit: Maximum consecutive retries when filters fail before terminating (default: 1000)
            allow_overlap: If False, samples cannot overlap with previously selected areas (default: True)

        Returns:
            Response containing task_id and collection name

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails
            TypeError: If extra filters have mismatched rate and resolution lists
            ValueError: If time_range or time_precision validation fails, or if try_limit is not positive
        """
        from datetime import datetime
        
        # Validate try_limit
        if try_limit < 1:
            raise ValueError(f"try_limit must be a positive integer, got {try_limit}")
        
        # Validate time_range and time_precision
        if (time_range is None) != (time_precision is None):
            raise ValueError("Both time_range and time_precision must be provided together, or both must be None")
        
        if time_precision is not None:
            valid_precisions = ['year', 'month', 'day', 'hour', 'minute', 'second']
            if time_precision not in valid_precisions:
                raise ValueError(f"time_precision must be one of {valid_precisions}, got '{time_precision}'")
        
        if time_range is not None:
            if not isinstance(time_range, (tuple, list)) or len(time_range) != 2:
                raise ValueError("time_range must be a tuple of (start_datetime, end_datetime)")
            
            # Validate ISO datetime format
            try:
                start_dt = datetime.fromisoformat(time_range[0].replace('Z', '+00:00'))
                end_dt = datetime.fromisoformat(time_range[1].replace('Z', '+00:00'))
                if start_dt >= end_dt:
                    raise ValueError(f"start_datetime ({time_range[0]}) must be before end_datetime ({time_range[1]})")
            except (ValueError, AttributeError) as e:
                raise ValueError(f"Invalid datetime format in time_range: {e}")
        
        expressions = [{"expr": expression_x, "res": res, "prefix": "x"}]
        
        res_y = res_y or res
        
        if expression_y != "skip":
            expressions.append({"expr": expression_y, "res": res_y, "prefix": "y"})
        
        filters = []
        if filter_x != "skip":
            filters.append({"expr": filter_x, "res": res, "rate": filter_x_rate})
        
        if filter_y != "skip":
            filters.append({"expr": filter_y, "res": res_y, "rate": filter_y_rate})
        
        if extra_filters:
            try:
                extra_filters_combined = zip(extra_filters, extra_filters_res, extra_filters_rate, strict=True)
            except TypeError:
                raise TypeError("Extra filters must have matching rate and resolution.")
            
            for expr, filter_res, rate in extra_filters_combined:
                filters.append({"expr": expr, "res": filter_res, "rate": rate})
        
        # Note: Template replacement for {year}, {month}, {day}, {hour}, etc. is now handled by the server
        # The server will randomly sample timestamps based on time_range and time_precision
        # and replace these placeholders for each sample
        
        if not skip_test:
            # For testing, we need to replace placeholders with valid test values
            # Use the start of the time_range as test values
            test_replacements = {}
            if time_range is not None:
                test_dt = datetime.fromisoformat(time_range[0].replace('Z', '+00:00'))
                test_replacements = {
                    "{year}": str(test_dt.year),
                    "{month}": str(test_dt.month),
                    "{day}": str(test_dt.day),
                    "{hour}": str(test_dt.hour),
                    "{minute}": str(test_dt.minute),
                    "{second}": str(test_dt.second)
                }
            
            for expr_dict in expressions:
                test_expr = expr_dict["expr"]
                for placeholder, value in test_replacements.items():
                    test_expr = test_expr.replace(placeholder, value)
                test_request = self._client.model._generate_test_request(test_expr, crs, -1)
                await self._client._terrakio_request("POST", "geoquery", json=test_request)
            
            for filter_dict in filters:
                test_expr = filter_dict["expr"]
                for placeholder, value in test_replacements.items():
                    test_expr = test_expr.replace(placeholder, value)
                test_request = self._client.model._generate_test_request(test_expr, crs, -1)
                await self._client._terrakio_request("POST", "geoquery", json=test_request)
        
        with open(aoi, 'r') as f:
            aoi_data = json.load(f)

        await self.get_collection(
            collection = collection,
        )

        payload = {
            "expressions": expressions,
            "filters": filters,
            "aoi": aoi_data,
            "samples": samples,
            "crs": crs,
            "tile_size": tile_size,
            "res": res,
            "output": "nc",
            "server": server,
            "try_limit": try_limit,
            "allow_overlap": allow_overlap
        }
        
        # Add time_range and time_precision if both are provided
        if time_range is not None and time_precision is not None:
            payload["time_range"] = list(time_range)  # Ensure it's a list for JSON serialization
            payload["time_precision"] = time_precision
        
        task_id_dict, status = await self._client._terrakio_request("POST", f"collections/{collection}/training_samples", json=payload)

        if status != 200:
            if status == 404:
                raise CollectionNotFoundError(f"Collection {collection} not found", status_code=status)
            raise GetTaskError(f"Training sample failed with status {status}", status_code=status)
        
        task_id = task_id_dict["task_id"]
        
        await self.track_progress(task_id)
        
        return {"task_id": task_id, "collection": collection}
            
    @require_api_key
    async def dataset(
        self,
        products: List[str],
        name: str,
        bucket: str = "terrakio",
        location: str = "testing/MSWXsmall",
        aoi: Optional[str] = None,
        expression: Optional[str] = None,
        date: Optional[str] = "2021-01-01",
        tile_size: float = 100,
        crs: str = "epsg:4326",
        res: float = 10,
        out_res: float = 10,
        no_data: float = -9999,
        dtype: str = "float32",
        create_doc: bool = False,
        skip_test: bool = False,
        force_res: bool = False,
        to_crs: Optional[str] = None,
        fully_cover: bool = True,
        skip_existing: bool = False,
    ) -> Dict[str, Any]:
        """
        Generate a dataset with the specified parameters.

        Args:
            products: List of product names
            name: Name of the dataset
            bucket: Storage bucket
            location: Storage location
            aoi: Path to GeoJSON file containing area of interest
            expression: Expression for data processing
            date: Date in YYYY-MM-DD format
            tile_size: Size of tiles (default: 100)
            crs: Coordinate reference system (default: "epsg:4326")
            res: Resolution (default: 10)
            out_res: Output resolution (default: 10)
            no_data: No data value (default: -9999)
            dtype: Data type (default: "float32")
            create_doc: Add dataset to the DB (default: False)
            skip_test: Skip testing the expression (default: False)
            force_res: Force resolution in case requests are too large (default: False)
            to_crs: Target coordinate reference system
            fully_cover: Fully cover the area (default: True)
            skip_existing: Skip existing data (default: False)

        Returns:
            Response containing task_id and collection name

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails
        """
        await self.create_collection(collection = name, bucket = bucket, location = location)
        date = _validate_date(date)
        sample = None
        reqs = []
        x_min, y_min, x_max, y_max, aoi_gdf = _get_bounds(aoi, crs, to_crs)
        if to_crs is None:
            to_crs = crs
        c=0
        for tile_req, i, j in _tile_generator(x_min, y_min, x_max, y_max, aoi_gdf, to_crs, res, tile_size, expression, OutputTypes.netcdf, fully_cover = fully_cover):
            c+=1
            if force_res:
                tile_req["force_res"] = True
            req_names = []
            for product in products:
                req_names.append(f"{product}_{date.strftime('%Y%m%d')}000000_{i:03d}_{j:03d}_00")
            reqs.append({"group": name, "file": req_names, "request": tile_req})
            if sample is None:
                sample = tile_req["expr"]
        i_max = int((x_max-x_min)/(tile_size*res))
        j_max = int((y_max-y_min)/(tile_size*res))
        geot = [x_min, out_res, 0, y_max, 0, -out_res]
        if not skip_test:
            result = await self._client.geoquery(**reqs[0]["request"], debug = "requests")
            request_count = result.get('request_count', 0)

        user_quota = await self._client.auth.get_user_quota()
        user_quota = user_quota.get('quota', -9999)
        
        if user_quota !=-9999 and user_quota < len(reqs) * request_count:
            raise QuotaInsufficientError(f"User quota is insufficient. Please contact support to increase your quota.")

        upload_urls = await self._get_upload_url(collection=name)
        url = upload_urls['url']
        await self._upload_json_data(reqs, url, use_gzip=True)

        payload = {"output": "snp", "skip_existing": skip_existing}
        
        task_id, status = await self._client._terrakio_request("POST", f"collections/{name}/generate_data", json=payload)
        task_id = task_id["task_id"]
        if dtype == Dataset_Dtype.uint8.value:
            no_data = int(no_data)
        if create_doc:
            await self._client.datasets.create_dataset(
                name=name,
                products=products,
                dates_iso8601=[date.isoformat()],
                proj4=pyproj.CRS.to_proj4(aoi_gdf.crs),
                i_max=i_max,
                j_max=j_max,
                x_size=int((res*tile_size)/out_res),
                y_size=int((res*tile_size)/out_res),
                geotransform=geot,
                no_data=no_data,
                data_type=dtype,
                bucket=bucket,
                path=f"{location}/%s_%s_%03d_%03d_%02d.snp",
                max_zoom=0,
            )

        await self.track_progress(task_id)

    @require_api_key
    async def tiles(
        self,
        collection: str,
        name: str = "irrigation_2019",
        aoi: Optional[str] = None,
        expression: str = "NSWIrrigation.landuse@(year=2019)",
        output: OutputTypes = OutputTypes.netcdf,
        tile_size: float = 10000,
        crs: str = "epsg:3577",
        res: float = 10,
        skip_test: bool = False,
        force_res: bool = False,
        to_crs: Optional[str] = None,
        fully_cover: bool = True,
        skip_existing: bool = False,
    ) -> Dict[str, Any]:
        """
        Generate tiles with the specified parameters.

        Args:
            collection: Name of the collection
            name: Name of the dataset (default: "irrigation_2019")
            aoi: Path to GeoJSON file containing area of interest
            expression: Expression for data processing (default: "NSWIrrigation.landuse@(year=2019)")
            output: Output format (default: "netcdf")
            tile_size: Size of tiles (default: 10000)
            crs: Coordinate reference system (default: "epsg:3577")
            res: Resolution (default: 10)
            skip_test: Skip testing the expression (default: False)
            force_res: Force resolution in case requests are too large (default: False)
            to_crs: Target coordinate reference system
            fully_cover: Fully cover the area (default: True)
            skip_existing: Skip existing data (default: False)

        Returns:
            Response containing task_id

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails
        """
        

        await self.get_collection(collection=collection)

        reqs = []
        sample = None
        x_min, y_min, x_max, y_max, aoi_gdf = _get_bounds(aoi, crs, to_crs)
        if to_crs is None:
            to_crs = crs
        for tile_req, i, j in _tile_generator(x_min, y_min, x_max, y_max, aoi_gdf, to_crs, res, tile_size, expression, output, fully_cover):
            if force_res:  
                tile_req["force_res"] = True
            req_name = f"{name}_{i:02d}_{j:02d}"
            reqs.append({"group": "tiles", "file": req_name, "request": tile_req})
            if sample is None:
                sample = tile_req["expr"]

        if not skip_test:
            result = await self._client.geoquery(**reqs[0]["request"], debug = "requests")
            request_count = result.get('request_count', 0)

        user_quota = await self._client.auth.get_user_quota()
        user_quota = user_quota.get('quota', -9999)
        
        if user_quota !=-9999 and user_quota < len(reqs) * request_count:
            raise QuotaInsufficientError(f"User quota is insufficient. Please contact support to increase your quota.")

        count = len(reqs)
        groups = list(set(dic["group"] for dic in reqs))
        self.console.print(f"[green]{count}[/green] requests with [blue]{len(groups)}[/blue] groups identified.")
        upload_urls = await self._get_upload_url(collection=collection)
        url = upload_urls['url']
        await self._upload_json_data(reqs, url, use_gzip=True)

        payload = {"output": output.value, "skip_existing": skip_existing}
        
        task_id, status = await self._client._terrakio_request("POST", f"collections/{collection}/generate_data", json=payload)
        task_id = task_id["task_id"]

        await self.track_progress(task_id)

    @require_api_key
    async def polygons(
        self,
        collection: str,
        aoi: str,
        expression: str = "mean:space(MSWX.air_temperature@(year=2022))",
        output: OutputTypes = OutputTypes.netcdf,
        id_field: str = "GID_0",
        crs: str = "epsg:4326",
        res: float = -1,
        skip_test: bool = False,
        skip_existing: bool = False,
    ) -> Dict[str, Any]:
        """
        Generate mass-stats for polygons in a GeoJSON file using the same expression.

        Args:
            collection: Name of the collection
            name: Name of the dataset
            aoi: Path to GeoJSON file containing area of interest
            expression: Expression for data processing (default: "mean:space(MSWX.air_temperature@(year=2022))")
            output: Output format (default: "netcdf")
            id_field: Field name to use as identifier (default: "GID_0")
            crs: Coordinate reference system (default: "epsg:4326")
            res: Resolution (default: -1)
            skip_test: Skip testing the expression (default: False)
            skip_existing: Skip existing data (default: False)

        Returns:
            Response containing task_id

        Raises:
            CollectionNotFoundError: If the collection is not found
            GetTaskError: If the API request fails
            ValueError: If id_field not found in feature properties or if id_field values are not unique
        """
        await self.get_collection(collection=collection)

        gdf = gpd.read_file(aoi)
        sample = None
        features = gdf.__geo_interface__
        reqs = []
        
        # Check if id_field exists and values are unique (fail-fast)
        seen = set()
        for feature in features["features"]:
            if id_field not in feature["properties"]:
                raise ValueError(f"ID field {id_field} not found in feature properties.")
            val = feature["properties"][id_field]
            if val in seen:
                raise ValueError(f"ID field '{id_field}' contains non-unique values. Please choose an id_field with unique values.")
            seen.add(val)
        
        for feature in features["features"]:
            feat = {
                "type": "Feature",
                "properties": {},
                "geometry": feature["geometry"],
            }
            request = {
                "feature": feat,
                "expr": expression,
                "output": output.value,
                "in_crs": crs, 
                "out_crs": crs,
                "resolution": res,
            }
            reqs.append({
                "group": "polygons", 
                "file": feature["properties"][id_field], 
                "request": request, 
                "metadata": _make_json_serializable(feature["properties"])
            })
            if sample is None:
                sample = request["expr"]

        # Test first request to ensure expression is valid
        if not skip_test:
            result = await self._client.geoquery(**reqs[0]["request"], debug="requests")
            request_count = result.get('request_count', 0)

        user_quota = await self._client.auth.get_user_quota()
        user_quota = user_quota.get('quota', -9999)
        
        if user_quota != -9999 and user_quota < len(reqs) * request_count:
            raise QuotaInsufficientError(f"User quota is insufficient. Please contact support to increase your quota.")
        
        count = len(reqs)
        groups = list(set(dic["group"] for dic in reqs))
        self.console.print(f"[green]{count}[/green] requests with [blue]{len(groups)}[/blue] groups identified.")
        upload_urls = await self._get_upload_url(collection=collection)
        url = upload_urls['url']
        await self._upload_json_data(reqs, url, use_gzip=True)

        payload = {"output": output.value, "skip_existing": skip_existing}
        
        task_id, status = await self._client._terrakio_request("POST", f"collections/{collection}/generate_data", json=payload)
        task_id = task_id["task_id"]

        await self.track_progress(task_id)

        return {"task_id": task_id, "collection": collection}




# Standard library imports
import asyncio
import psutil
import random
import uuid
from io import BytesIO
from typing import Optional

# Third-party library imports
import aiohttp
import geopandas as gpd
import pandas as pd
import pyproj
import xarray as xr
from geopandas import GeoDataFrame
from shapely.geometry import box, mapping, shape
from shapely.ops import transform
import threading
from concurrent.futures import ThreadPoolExecutor

# Local imports
from .geoquries import request_geoquery_list

class cloud_object(gpd.GeoDataFrame):
    """
    This class is a class used for cloud
    """
    def __init__(self, job_id: str, job_name: str, client=None):

        super().__init__({
            'geometry': [], 
            'dataset': []
        })

        self.job_id = job_id
        self.client = client
        self.job_name = job_name

    def __repr__(self):
        return (
            f"<CloudZonalStats job_id='{self.job_id}', job_name='{self.job_name}'>\n"
            f"Call .head(n) to fetch a preview GeoDataFrame when the job completes."
        )

    def _repr_html_(self):
        # Jupyter HTML-friendly representation to avoid auto-rendering an empty DataFrame
        return (
            f"<div style='font-family:system-ui,Segoe UI,Helvetica,Arial,sans-serif'>"
            f"<strong>Cloud Zonal Stats</strong><br/>"
            f"job_id: <code>{self.job_id}</code><br/>"
            f"job_name: <code>{self.job_name}</code><br/>"
            f"<em>Use <code>.head(n)</code> to retrieve a preview once the job is completed.</em>"
            f"</div>"
        )

    def head(self, n = 5):
        """
        Returns the first n files stored in the cloud bucket.
        """
        # Detect if we're inside an existing event loop (e.g., Jupyter)
        in_running_loop = False
        try:
            asyncio.get_running_loop()
            in_running_loop = True
        except RuntimeError:
            in_running_loop = False

        if in_running_loop:
            # Run the async function in a separate thread with its own loop
            def run_async_in_thread():
                new_loop = asyncio.new_event_loop()
                try:
                    return new_loop.run_until_complete(self._head_async(n))
                finally:
                    new_loop.close()
            
            with ThreadPoolExecutor(max_workers=1) as executor:
                future = executor.submit(run_async_in_thread)
                return future.result()
        else:
            # No running loop - safe to use asyncio.run
            return asyncio.run(self._head_async(n))
        
    async def _head_async(self, n = 5):
        """
        Returns the first n files stored in the cloud bucket.

        Args:
            n (int): Number of files to return. Default is 5.

        Returns:
            GeoDataFrame: A GeoDataFrame containing the first n files.
        """

        track_info = await self.client.collections.track_job([self.job_id])
        job_info = track_info[self.job_id]
        status = job_info['status']
        
        if status == "Completed":
            payload = {
                "job_name": job_info["name"],
                "file_type": "raw",
                "bucket": job_info["bucket"],
            }
            result = await self.client._terrakio_request("POST", "mass_stats/download_files", json=payload)
            download_urls = result["download_urls"][:n]
            datasets = []

            async with aiohttp.ClientSession() as session:
                for i, url in enumerate(download_urls):
                    try:
                        self.client.logger.info(f"Downloading dataset {i+1}/{len(download_urls)}...")
                        async with session.get(url) as response:
                            if response.status == 200:
                                content = await response.read()
                                dataset = xr.open_dataset(BytesIO(content))
                                datasets.append(dataset)
                                self.client.logger.info(f"Successfully processed dataset {i+1}")
                            else:
                                self.client.logger.warning(f"Failed to download dataset {i+1}: HTTP {response.status}")
                    except Exception as e:
                        self.client.logger.error(f"Error downloading dataset {i+1}: {e}")
                        continue
                if not datasets:
                    self.client.logger.warning("No datasets were successfully downloaded")
                    return gpd.GeoDataFrame({'geometry': [], 'dataset': []})
                try:
                    json_response = await self.client._terrakio_request(
                        "POST", "mass_stats/download_json", 
                        params={"job_name": job_info['name']}
                    )
                    json_url = json_response["download_url"]
                    
                    async with session.get(json_url) as response:
                        if response.status == 200:
                            json_data = await response.json()
                            self.client.logger.info("Successfully downloaded geometry data")
                            
                            geometries = []
                            max_geometries = min(n, len(json_data), len(datasets))
                            
                            for i in range(max_geometries):
                                try:
                                    geom_dict = json_data[i]["request"]["feature"]["geometry"]
                                    shapely_geom = shape(geom_dict)
                                    geometries.append(shapely_geom)
                                except (KeyError, ValueError) as e:
                                    self.client.logger.warning(f"Error parsing geometry {i}: {e}")
                                    continue
                            
                            min_length = min(len(datasets), len(geometries))
                            if min_length == 0:
                                self.client.logger.warning("No matching datasets and geometries found")
                                return gpd.GeoDataFrame({'geometry': [], 'dataset': []})
                            
                            gdf = gpd.GeoDataFrame({
                                'geometry': geometries[:min_length],
                                'dataset': datasets[:min_length]
                            })
                            
                            self.client.logger.info(f"Created GeoDataFrame with {len(gdf)} rows")

                            # Derive id values from json metadata (prefer 'file', fallback to 'group')
                            id_values = []
                            for i in range(min_length):
                                entry = json_data[i] if i < len(json_data) else {}
                                id_candidate = entry.get('file') or entry.get('group') or ''
                                if isinstance(id_candidate, str) and id_candidate.startswith('file_'):
                                    id_val = id_candidate[len('file_'):]
                                elif isinstance(id_candidate, str) and id_candidate.startswith('group_'):
                                    id_val = id_candidate[len('group_'):]
                                else:
                                    id_val = str(id_candidate) if id_candidate else str(i)
                                id_values.append(id_val)

                            # Geometry to id mapping using WKB to avoid precision issues
                            geom_to_id = {geometries[i].wkb: id_values[i] for i in range(min_length)}

                            try:
                                expanded_gdf = expand_on_variables_and_time(gdf)

                                # Attach id as first index level, geometry second, time third if present
                                if hasattr(expanded_gdf.index, 'names') and 'geometry' in expanded_gdf.index.names:
                                    if isinstance(expanded_gdf.index, pd.MultiIndex):
                                        geometry_index = expanded_gdf.index.get_level_values('geometry')
                                    else:
                                        geometry_index = expanded_gdf.index
                                    id_col = [geom_to_id.get(geom.wkb) for geom in geometry_index]
                                    expanded_gdf['id'] = id_col
                                    expanded_gdf = expanded_gdf.reset_index()
                                    if 'time' in expanded_gdf.columns:
                                        expanded_gdf = expanded_gdf.set_index(['id', 'geometry', 'time'])
                                    else:
                                        expanded_gdf = expanded_gdf.set_index(['id', 'geometry'])
                                else:
                                    # geometry exists as a column
                                    id_col = [geom_to_id.get(geom.wkb) for geom in expanded_gdf['geometry']]
                                    expanded_gdf['id'] = id_col
                                    if 'time' in expanded_gdf.columns:
                                        expanded_gdf = expanded_gdf.set_index(['id', 'geometry', 'time'])
                                    else:
                                        expanded_gdf = expanded_gdf.set_index(['id', 'geometry'])

                                return expanded_gdf
                            except NameError:
                                self.client.logger.warning("expand_on_variables_and_time function not found, returning raw GeoDataFrame")
                                # Set id on raw gdf and index appropriately
                                gdf['id'] = id_values
                                return gdf.set_index(['id', 'geometry'])
                                
                        else:
                            self.client.logger.warning(f"Failed to download geometry data: HTTP {response.status}")
                            return gpd.GeoDataFrame({'geometry': [], 'dataset': []})
                                    
                except Exception as e:
                        self.client.logger.error(f"Error downloading geometry data: {e}")
                        return gpd.GeoDataFrame({'geometry': [], 'dataset': []})
        
        elif status in ["Failed", "Cancelled", "Error"]:
            raise RuntimeError(f"The zonal stats job (job_id: {self.job_id}) has failed, cancelled, or errored. Please check the job status!")
        
        else:
            # Job is still running - include progress information
            completed = job_info.get('completed', 0)
            total = job_info.get('total', 1)
            progress = completed / total if total > 0 else 0
            percentage = progress * 100
            
            # Create progress bar
            bar_length = 30  # Shorter bar for error message
            filled_length = int(bar_length * progress)
            bar = '█' * filled_length + '░' * (bar_length - filled_length)
            
            raise RuntimeError(
                f"The zonal stats job (job_id: {self.job_id}) is still running. "
                f"Progress: [{bar}] {percentage:.1f}% ({completed}/{total}). "
                f"Please come back at a later time!"
            )
    
def expand_on_time(gdf):
    """
    Expand datasets on time dimension - each time becomes a new row.
    
    Input: GeoDataFrame with 'geometry' and 'dataset' columns (or variable columns)
    Output: GeoDataFrame with time in multi-index and datasets without time coordinate
    """
    rows = []
    
    for idx, row in gdf.iterrows():
        if 'geometry' in gdf.columns:
            geometry = row['geometry']
        elif gdf.index.name == 'geometry':
            geometry = idx
        else:
            raise ValueError(f"Cannot find geometry in columns: {list(gdf.columns)} or index: {gdf.index.name}")
        
        if 'dataset' in gdf.columns:
            dataset = row['dataset']
            
            if 'time' in dataset.dims:
                for time_val in dataset.time.values:
                    time_slice = dataset.sel(time=time_val).drop_vars('time')
                    rows.append({
                        'geometry': geometry,
                        'time': time_val,
                        'dataset': time_slice
                    })
            else:
                rows.append({
                    'geometry': geometry,
                    'dataset': dataset
                })
        else:
            variable_columns = list(gdf.columns)
            
            first_dataset = row[variable_columns[0]]
            if 'time' in first_dataset.dims:
                time_values = first_dataset.time.values
                
                for time_val in time_values:
                    row_data = {'geometry': geometry, 'time': time_val}
                    
                    for var_col in variable_columns:
                        dataset = row[var_col]
                        time_slice = dataset.sel(time=time_val).drop_vars('time')
                        row_data[var_col] = time_slice
                    
                    rows.append(row_data)
            else:
                row_data = {'geometry': geometry}
                for var_col in variable_columns:
                    row_data[var_col] = row[var_col]
                rows.append(row_data)
    
    result_df = pd.DataFrame(rows)
    
    if 'time' in result_df.columns:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry', 'time'])
    else:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry'])
    
    result_gdf.attrs = gdf.attrs.copy()
    
    return result_gdf

def expand_on_variables(gdf):
    """
    Expand datasets on variables dimension - each variable becomes a new column.
    
    Input: GeoDataFrame with 'geometry' and 'dataset' columns (or already time-expanded)
    Output: GeoDataFrame with separate column for each variable
    """
    rows = []
    
    for idx, row in gdf.iterrows():
        if 'geometry' in gdf.columns:
            geometry = row['geometry']
        elif hasattr(gdf.index, 'names') and 'geometry' in gdf.index.names:
            if isinstance(idx, tuple):
                geometry_idx = gdf.index.names.index('geometry')
                geometry = idx[geometry_idx]
                time_idx = gdf.index.names.index('time')
                time_val = idx[time_idx]
            else:
                geometry = idx
                time_val = None
        else:
            raise ValueError(f"Cannot find geometry in columns: {list(gdf.columns)} or index: {gdf.index.names}")
        
        if 'dataset' in gdf.columns:
            dataset = row['dataset']
            
            var_names = list(dataset.data_vars.keys())
            
            if len(var_names) <= 1:
                if len(var_names) == 0:
                    continue
            
            if hasattr(gdf.index, 'names') and 'time' in gdf.index.names:
                row_data = {'geometry': geometry, 'time': time_val}
            else:
                row_data = {'geometry': geometry}
            
            for var_name in var_names:
                var_dataset = dataset[[var_name]]
                
                if len(var_dataset.dims) == 0:
                    row_data[var_name] = float(var_dataset[var_name].values)
                else:
                    row_data[var_name] = var_dataset
            
            rows.append(row_data)
        else:
            raise ValueError("Expected 'dataset' column for variable expansion")
    
    result_df = pd.DataFrame(rows)

    if 'time' in result_df.columns:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry', 'time'])
    else:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry'])
    
    result_gdf.attrs = gdf.attrs.copy()
    
    return result_gdf

def expand_on_variables_and_time(gdf):
    """
    Convenience function to expand on both variables and time.
    Automatically detects which expansions are possible.
    """
    try:
        expanded_on_time = expand_on_time(gdf)
    except Exception as e:
        expanded_on_time = gdf
    
    try:
        expanded_on_variables_and_time = expand_on_variables(expanded_on_time)
        return expanded_on_variables_and_time
    except Exception as e:
        return expanded_on_time
    
def estimate_geometry_size_ratio(queries: list):
    """Calculate size ratios for all geometries relative to the first geometry using bounding box area."""
    
    areas = []
    
    for query in queries:
        geom = shape(query["feature"]["geometry"])
        in_crs = query["in_crs"]
        
        if in_crs and in_crs != 'EPSG:3857':
            transformer = pyproj.Transformer.from_crs(in_crs, 'EPSG:3857', always_xy=True)
            transformed_geom = transform(transformer.transform, geom)
            bbox = box(*transformed_geom.bounds)
            area = bbox.area
        else:
            bbox = box(*geom.bounds)
            area = bbox.area
        
        areas.append(area)    
    base_area = areas[0]
    
    if base_area == 0:
        non_zero_areas = [area for area in areas if area > 0]
        base_area = non_zero_areas[0] if non_zero_areas else 1.0
    
    ratios = []
    for area in areas:
        if area == 0:
            ratios.append(0.1)
        else:
            ratios.append(area / base_area)
    
    return ratios

async def estimate_query_size(
    client,
    quries: list[dict],
):
    first_query = quries[0]

    first_query_dataset = await client.geoquery(**first_query)
    ratios = estimate_geometry_size_ratio(quries)
    total_size_mb = 0
    for i in range(len(ratios)):
        total_size_mb += first_query_dataset.nbytes * ratios[i] / (1024**2)
    return total_size_mb

async def estimate_timestamp_number(
        client,
        quries: list[dict],
):
    if len(quries) <= 3:
        return quries
    sampled_queries = [query.copy() for query in random.sample(quries, 3)]
    for query in sampled_queries:
        query['debug'] = 'grpc'
    result = await request_geoquery_list(client = client, quries = sampled_queries, conc = 5)
    total_estimated_number_of_timestamps = result * len(quries)
    return total_estimated_number_of_timestamps


def get_available_memory_mb():
    """
    Get available system memory in MB
    
    Returns:
        float: Available memory in MB
    """
    memory = psutil.virtual_memory()
    available_mb = memory.available / (1024 * 1024)
    return round(available_mb, 2)

async def local_or_remote(
        client,
        quries: list[dict],
):
    if len(quries) > 1000:
        return {
            "local_or_remote": "remote",
            "reason": "The number of the requests is too large(>1000), please set the mass_stats parameter to True",
        }
    elif await estimate_timestamp_number(client = client, quries = quries) > 25000:
        return {
            "local_or_remote": "remote",
            "reason": "The time taking for making these requests is too long, please set the mass_stats parameter to True",
        }
    elif await estimate_query_size(client = client, quries = quries) > get_available_memory_mb():
        return {
            "local_or_remote": "remote",
            "reason": "The size of the dataset is too large, please set the mass_stats parameter to True",
        }
    else:
        return {
            "local_or_remote": "local",
            "reason": "The number of the requests is not too large, and the time taking for making these requests is not too long, and the size of the dataset is not too large",
        }
    
def gdf_to_json(
    gdf: GeoDataFrame,
    expr: str,
    in_crs: str = "epsg:4326",
    out_crs: str = "epsg:4326",
    resolution: int = -1,
    geom_fix: bool = False,
    id_column: Optional[str] = None,
):
    """
    Convert a GeoDataFrame to a list of JSON requests for collections processing.
    
    Args:
        gdf: GeoDataFrame containing geometries and optional metadata
        expr: Expression to evaluate
        in_crs: Input coordinate reference system
        out_crs: Output coordinate reference system
        resolution: Resolution parameter
        geom_fix: Whether to fix geometry issues
        id_column: Optional column name to use for group and file names
        
    Returns:
        list: List of dictionaries formatted for collections requests
    """
    collections_requests = []
    
    for idx, row in gdf.iterrows():
        request_feature = {
            "expr": expr,
            "feature": {
                "type": "Feature",
                "geometry": mapping(gdf.geometry.iloc[idx]),
                "properties": {}
            },
            "in_crs": in_crs,
            "out_crs": out_crs,
            "resolution": resolution,
            "geom_fix": geom_fix,
        }
        
        if id_column is not None and id_column in gdf.columns:
            identifier = str(row[id_column])
            group_name = f"group_{identifier}"
            file_name = f"file_{identifier}"
        else:
            group_name = f"group_{idx}"
            file_name = f"file_{idx}"
            
        request_entry = {
            "group": group_name,
            "file": file_name,
            "request": request_feature,
        }
        
        collections_requests.append(request_entry)
        
    return collections_requests

async def handle_collections(
    client,
    gdf: GeoDataFrame,
    expr: str,
    in_crs: str = "epsg:4326",
    out_crs: str = "epsg:4326",
    resolution: int = -1,
    geom_fix: bool = False,
    id_column: Optional[str] = None,
):
    request_json = gdf_to_json(gdf=gdf, expr=expr, in_crs=in_crs, out_crs=out_crs, 
                              resolution=resolution, geom_fix=geom_fix, id_column=id_column)
    
    job_response = await client.collections.execute_job(
        name=f"zonal-stats-{str(uuid.uuid4())[:6]}",
        output="netcdf",
        config={},
        request_json=request_json,
        overwrite=True,
    )
    
    # Extract the actual task ID from the response
    if isinstance(job_response, dict) and 'task_id' in job_response:
        return job_response['task_id']  # Return just the string ID
    else:
        return job_response  # In case it's already just the ID


async def zonal_stats(
    client,
    gdf: GeoDataFrame,
    expr: str,
    conc: int = 20,
    in_crs: str = "epsg:4326",
    out_crs: str = "epsg:4326",
    resolution: int = -1,
    geom_fix: bool = False,
    mass_stats: bool = False,
    id_column: Optional[str] = None,
):
    """Compute zonal statistics for all geometries in a GeoDataFrame."""
    if mass_stats:
        collections_id = await handle_collections(
            client = client,
            gdf = gdf,
            expr = expr,
            in_crs = in_crs,
            out_crs = out_crs,
            resolution = resolution,
            geom_fix = geom_fix,
            id_column = id_column,
        )
        job_name = await client.collections.track_job([collections_id])
        job_name = job_name[collections_id]["name"]
        cloud_files_object = cloud_object(job_id = collections_id, job_name = job_name, client = client)

        return cloud_files_object
    
    quries = []
    for i in range(len(gdf)):
        quries.append({
            "expr": expr,
            "feature": {
                "type": "Feature",
                "geometry": mapping(gdf.geometry.iloc[i]),
                "properties": {}
            },
            "in_crs": in_crs,
            "out_crs": out_crs,
            "resolution": resolution,
            "geom_fix": geom_fix,
        })

    local_or_remote_result = await local_or_remote(client= client, quries = quries)
    if local_or_remote_result["local_or_remote"] == "remote":
        raise ValueError(local_or_remote_result["reason"])
    else:
        gdf_with_datasets = await request_geoquery_list(client = client, quries = quries, conc = conc)
        gdf_with_datasets.attrs["cloud_metadata"] = {
            "is_cloud_backed": False,
        } 
        gdf_with_datasets = expand_on_variables_and_time(gdf_with_datasets)
        
        # If an id_column is provided, attach it to the result and include in the index
        if id_column is not None and id_column in gdf.columns:
            # Build a mapping from input geometries to id values (use WKB for robust equality)
            geometry_to_id = {geom.wkb: id_val for geom, id_val in zip(gdf.geometry, gdf[id_column])}

            # Determine geometry values in the result (index may be geometry or (geometry, time))
            if hasattr(gdf_with_datasets.index, 'names') and 'geometry' in gdf_with_datasets.index.names:
                if isinstance(gdf_with_datasets.index, pd.MultiIndex):
                    geometry_index = gdf_with_datasets.index.get_level_values('geometry')
                else:
                    geometry_index = gdf_with_datasets.index
                id_values = [geometry_to_id.get(geom.wkb) for geom in geometry_index]
                gdf_with_datasets[id_column] = id_values
                # Reset index to control index composition precisely, then set to desired levels
                gdf_with_datasets = gdf_with_datasets.reset_index()
                if 'time' in gdf_with_datasets.columns:
                    gdf_with_datasets = gdf_with_datasets.set_index([id_column, 'geometry', 'time'])
                else:
                    gdf_with_datasets = gdf_with_datasets.set_index([id_column, 'geometry'])
            else:
                # geometry exists as a column
                id_values = [geometry_to_id.get(geom.wkb) for geom in gdf_with_datasets['geometry']]
                gdf_with_datasets[id_column] = id_values
                if 'time' in gdf_with_datasets.columns:
                    gdf_with_datasets = gdf_with_datasets.set_index([id_column, 'geometry', 'time'])
                else:
                    gdf_with_datasets = gdf_with_datasets.set_index([id_column, 'geometry'])
    return gdf_with_datasets




from typing import Any, Dict, List, Optional

from ..exceptions import (
    CommandPermissionError,
    DatasetNotFoundError,
    DatasetPermissionError,
    GetDatasetError,
    ListDatasetsError,
    CreateDatasetError,
    DatasetAlreadyExistsError,
    DeleteDatasetError,
    OverwriteDatasetError,
)
from ..helper.decorators import require_api_key, require_auth, require_token

class DatasetManagement:
    def __init__(self, client):
        self._client = client

    @require_api_key
    async def list_datasets(self, substring: Optional[str] = None) -> List[Dict[str, Any]]:
        """
        List datasets, optionally filtering by a substring and collection.
        
        Args:
            substring: Substring to filter by (optional)
            
        Returns:
            List of datasets matching the criteria
        """
        params = {"substring": substring} if substring else None
        response, status = await self._client._terrakio_request("GET", "/datasets", params=params)
        if status != 200:
            raise ListDatasetsError(f"List datasets failed with status {status}", status_code=status)
        return response

    @require_api_key
    async def get_dataset(self, name: str) -> Dict[str, Any]:
        """
        Retrieve dataset info by dataset name.
        
        Args:
            name: The name of the dataset (required)
            
        Returns:
            Dataset information as a dictionary
            
        Raises:
            GetDatasetError: If the API request fails
            DatasetNotFoundError: If the dataset is not found
            DatasetPermissionError: If the user does not have permission to get the dataset
        """
        response, status = await self._client._terrakio_request("GET", f"/datasets/{name}")
        if status != 200:
            if status == 404:
                raise DatasetNotFoundError(f"Dataset {name} not found", status_code = status)
            if status == 403:
                raise DatasetPermissionError(f"You do not have permission to get dataset {name}", status_code = status)
            raise GetDatasetError(f"Get dataset failed with status {status}", status_code = status)
        return response

    @require_api_key
    async def create_dataset(
        self, 
        name: str, 
        products: Optional[List[str]] = None,
        dates_iso8601: Optional[List[str]] = None,
        bucket: Optional[str] = None,
        path: Optional[str] = None,
        data_type: Optional[str] = None,
        no_data: Optional[Any] = None,
        i_max: Optional[int] = None,
        j_max: Optional[int] = None,
        y_size: Optional[int] = None,
        x_size: Optional[int] = None,
        proj4: Optional[str] = None,
        abstract: Optional[str] = None,
        geotransform: Optional[List[float]] = None,
        padding: Optional[Any] = None,
        input: Optional[str] = None,
        max_zoom: Optional[int] = None
    ) -> Dict[str, Any]:
        """
        Create a new dataset.
                            
        Args:
            name: Name of the dataset (required)
            products: List of products
            dates_iso8601: List of dates (will be automatically sorted chronologically)
            bucket: Storage bucket
            path: Storage path
            data_type: Data type
            no_data: No data value
            i_max: Maximum level
            j_max: Maximum level
            y_size: Y size
            x_size: X size
            proj4: Projection string
            abstract: Dataset abstract
            geotransform: Geotransform parameters
            padding: Padding value
            max_zoom: Maximum zoom level
                                
        Returns:
            Created dataset information
                                
        Raises:
            APIError: If the API request fails
        """
        payload = {"name": name}
        param_mapping = {
            "products": products,
            "dates_iso8601": dates_iso8601,
            "bucket": bucket,
            "path": path,
            "data_type": data_type,
            "no_data": no_data,
            "i_max": i_max,
            "j_max": j_max,
            "y_size": y_size,
            "x_size": x_size,
            "proj4": proj4,
            "abstract": abstract,
            "geotransform": geotransform,
            "padding": padding,
            "input": input,
            "max_zoom": max_zoom
        }
        for param, value in param_mapping.items():
            if value is not None:
                payload[param] = value
        response, status = await self._client._terrakio_request("POST", "/datasets", json = payload)
        if status != 200:
            if status ==403:
                raise CommandPermissionError(f"You do not have permission to create dataset {name}", status_code = status)
            elif status == 409:
                raise DatasetAlreadyExistsError(f"Dataset {name} already exists", status_code = status)
            raise CreateDatasetError(f"Create dataset failed with status {status}", status_code = status)
        return response
        
    @require_api_key
    async def delete_dataset(self, name: str) -> Dict[str, Any]:
        """
        Delete a dataset by name.
        
        Args:
            name: The name of the dataset (required)
            
        Returns:
            Deleted dataset information
            
        Raises:
            CommandPermissionError: If the user does not have permission to delete the dataset
            DeleteDatasetError: If the API request fails
        """
        response, status = await self._client._terrakio_request("DELETE", f"/datasets/{name}")
        if status != 200:
            if status == 403:
                raise CommandPermissionError(f"You do not have permission to delete dataset {name}", status_code = status)
            raise DeleteDatasetError(f"Delete dataset failed with status {status}", status_code = status)
        return response

    @require_api_key
    def update_dataset(
        self, 
        name: str, 
        append: bool = True, 
        products: Optional[List[str]] = None,
        dates_iso8601: Optional[List[str]] = None,
        bucket: Optional[str] = None,
        path: Optional[str] = None,
        data_type: Optional[str] = None,
        no_data: Optional[Any] = None,
        i_max: Optional[int] = None,
        j_max: Optional[int] = None,
        y_size: Optional[int] = None,
        x_size: Optional[int] = None,
        proj4: Optional[str] = None,
        abstract: Optional[str] = None,
        geotransform: Optional[List[float]] = None,
        padding: Optional[Any] = None
    ) -> Dict[str, Any]:
        """
        Update an existing dataset.
                            
        Args:
            name: Name of the dataset (required)
            append: Whether to append data or replace (default: True)
            products: List of products
            dates_iso8601: List of dates (will be automatically sorted chronologically)
            bucket: Storage bucket
            path: Storage path
            data_type: Data type
            no_data: No data value
            i_max: Maximum level
            j_max: Maximum level
            y_size: Y size
            x_size: X size
            proj4: Projection string
            abstract: Dataset abstract
            geotransform: Geotransform parameters
            padding: Padding value
                                
        Returns:
            Updated dataset information
                                
        Raises:
            APIError: If the API request fails
        """
        if dates_iso8601 is not None:
            dates_iso8601 = sorted(dates_iso8601)

        params = {"append": str(append).lower()}
        payload = {"name": name}
        param_mapping = {
            "products": products,
            "dates_iso8601": dates_iso8601,
            "bucket": bucket,
            "path": path,
            "data_type": data_type,
            "no_data": no_data,
            "i_max": i_max,
            "j_max": j_max,
            "y_size": y_size,
            "x_size": x_size,
            "proj4": proj4,
            "abstract": abstract,
            "geotransform": geotransform,
            "padding": padding
        }
        for param, value in param_mapping.items():
            if value is not None:
                payload[param] = value
        return self._client._terrakio_request("PATCH", "/datasets", params = params, json = payload)
    
    @require_api_key
    def update_virtual_dataset(
        self, 
        name: str, 
        append: bool = True, 
        products: Optional[List[str]] = None,
        dates_iso8601: Optional[List[str]] = None,
        bucket: Optional[str] = None,
        path: Optional[str] = None,
        data_type: Optional[str] = None,
        no_data: Optional[Any] = None,
        i_max: Optional[int] = None,
        j_max: Optional[int] = None,
        y_size: Optional[int] = None,
        x_size: Optional[int] = None,
        proj4: Optional[str] = None,
        abstract: Optional[str] = None,
        geotransform: Optional[List[float]] = None,
        padding: Optional[Any] = None,
        input: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Update an existing dataset.
                            
        Args:
            name: Name of the dataset (required)
            append: Whether to append data or replace (default: True)
            products: List of products
            dates_iso8601: List of dates (will be automatically sorted chronologically)
            bucket: Storage bucket
            path: Storage path
            data_type: Data type
            no_data: No data value
            i_max: Maximum level
            j_max: Maximum level
            y_size: Y size
            x_size: X size
            proj4: Projection string
            abstract: Dataset abstract
            geotransform: Geotransform parameters
            padding: Padding value
            input: The input for the virtual dataset
                                
        Returns:
            Updated dataset information
                                
        Raises:
            APIError: If the API request fails
        """
        # Sort dates_iso8601 chronologically if provided
        if dates_iso8601 is not None:
            dates_iso8601 = sorted(dates_iso8601)

        params = {"append": str(append).lower()}
        payload = {"name": name}
        param_mapping = {
            "products": products,
            "dates_iso8601": dates_iso8601,
            "bucket": bucket,
            "path": path,
            "data_type": data_type,
            "no_data": no_data,
            "i_max": i_max,
            "j_max": j_max,
            "y_size": y_size,
            "x_size": x_size,
            "proj4": proj4,
            "abstract": abstract,
            "geotransform": geotransform,
            "padding": padding,
            "input": input,
        }
        for param, value in param_mapping.items():
            if value is not None:
                payload[param] = value
        return self._client._terrakio_request("PATCH", "/datasets", params = params, json = payload)


    @require_api_key
    async def _get_url_for_upload_inference_script(self, script_path: str) -> str:
        """
        Get the url for the upload of the inference script
        """
        # we have the path, and we just need to get the bucket name
        payload = {
            "script_path": script_path,
        }
        return await self._client._terrakio_request("POST", "models/update_inference_script", json = payload)
    
    @require_api_key
    async def update_virtual_dataset_inference(
        self,
        name: str,
        inference_script_path: str,
        append: bool = True,
    ):
        """
        Update the inference script for a virtual dataset.
        """
        params = {"append": str(append).lower()}
        dataset_info = await self.get_dataset(name)
        print("the current dataset info is: ", dataset_info)
        script_path = dataset_info["path"]
        product_name = dataset_info["products"][0]
        script_path = script_path + f"/{product_name}.py"
        upload_url_dict = await self._get_url_for_upload_inference_script(script_path)
        upload_url = upload_url_dict["script_upload_url"]
        try:
            with open(inference_script_path, "r", encoding="utf-8") as f:
                script_content = f.read()
                script_bytes = script_content.encode('utf-8')
                headers = {
                    "Content-Type": "text/x-python",
                    "Content-Length": str(len(script_bytes))
                }
                response = await self._client._regular_request("PUT", endpoint=upload_url, data=script_bytes, headers=headers)
        except FileNotFoundError:
            raise FileNotFoundError(f"Inference script file not found: {inference_script_path}")
        except Exception as e:
            raise Exception(f"Failed to upload inference script: {str(e)}")
    
    @require_api_key
    async def overwrite_dataset(
        self, 
        name: str, 
        products: Optional[List[str]] = None,
        dates_iso8601: Optional[List[str]] = None,
        bucket: Optional[str] = None,
        path: Optional[str] = None,
        data_type: Optional[str] = None,
        no_data: Optional[Any] = None,
        i_max: Optional[int] = None,
        j_max: Optional[int] = None,
        y_size: Optional[int] = None,
        x_size: Optional[int] = None,
        proj4: Optional[str] = None,
        abstract: Optional[str] = None,
        geotransform: Optional[List[float]] = None,
        padding: Optional[Any] = None
    ) -> Dict[str, Any]:
        """
        Overwrite a dataset.
                            
        Args:
            name: Name of the dataset (required)
            products: List of products
            dates_iso8601: List of dates (will be automatically sorted chronologically)
            bucket: Storage bucket
            path: Storage path
            data_type: Data type
            no_data: No data value
            i_max: Maximum level
            j_max: Maximum level
            y_size: Y size
            x_size: X size
            proj4: Projection string
            abstract: Dataset abstract
            geotransform: Geotransform parameters
            padding: Padding value
                                
        Returns:
            Overwritten dataset information
                                
        Raises:
            CommandPermissionError: If the user does not have permission to overwrite the dataset
            DatasetNotFoundError: If the dataset is not found
            OverwriteDatasetError: If the API request fails
        """
        payload = {"name": name}
        param_mapping = {
            "products": products,
            "dates_iso8601": dates_iso8601,
            "bucket": bucket,
            "path": path,
            "data_type": data_type,
            "no_data": no_data,
            "i_max": i_max,
            "j_max": j_max,
            "y_size": y_size,
            "x_size": x_size,
            "proj4": proj4,
            "abstract": abstract,
            "geotransform": geotransform,
            "padding": padding
        }
        for param, value in param_mapping.items():
            if value is not None:
                payload[param] = value
        response, status = await self._client._terrakio_request("PUT", "/datasets", json = payload)
        if status != 200:
            if status == 403:
                raise CommandPermissionError(f"You do not have permission to overwrite dataset {name}", status_code = status)
            elif status == 404:
                raise DatasetNotFoundError(f"Dataset {name} not found", status_code = status)
            raise OverwriteDatasetError(f"Failed to overwrite dataset: {response}", status_code = status)
        return response

    @require_api_key
    def download_file_to_path(self, job_name, stage, file_name, output_path):
        if not self.collections:
            from terrakio_core.endpoints.collections import Collections
            if not self.url or not self.key:
                raise ConfigurationError("Collections client not initialized. Make sure API URL and key are set.")
            self.collections = Collections(
                base_url=self.url,
                api_key=self.key,
                verify=self.verify,
                timeout=self.timeout
            )
        taskid = self.collections.get_task_id(job_name, stage).get('task_id')
        trackinfo = self.collections.track_job([taskid])
        bucket = trackinfo[taskid]['bucket']
        return self.collections.download_file(job_name, bucket, file_name, output_path)


import asyncio
import os
import tempfile
import time
import uuid

from ..helper.tiles import tiles

async def create_dataset_file(
    client,
    aoi: str,
    expression: str,
    output: str,
    download_path: str,
    in_crs: str = "epsg:4326",
    to_crs: str = "epsg:4326",
    res: float = 0.0001,
    region: str = None,
    overwrite: bool = False,
    skip_existing: bool = False,
    non_interactive: bool = True,
    name: str | None = None,
    poll_interval: int = 30,
    max_file_size_mb: int = 5120,
    tile_size: int = 1024,
    mask: bool = True
) -> dict:
    
    if not name:
        name = f"file-gen-{uuid.uuid4().hex[:8]}"
    
    body, reqs, groups = tiles(
        name = name, 
        aoi = aoi, 
        expression = expression,
        output = output,
        tile_size = tile_size,
        crs = in_crs,
        res = res,
        region = region,
        to_crs = to_crs,
        mask = mask,
        overwrite = overwrite,
        skip_existing = skip_existing,
        non_interactive = non_interactive
    )
    with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as tempreq:
        tempreq.write(reqs)
        tempreqname = tempreq.name

    task_id = await client.collections.execute_job(
        name=body["name"],
        region=body["region"],
        output=body["output"],
        config = {},
        overwrite=body["overwrite"],
        skip_existing=body["skip_existing"],
        request_json=tempreqname,
    )

    start_time = time.time()
    status = None
    client.logger.info(f"Tracking data generation job {task_id['task_id']}...")
    while True:
        try:
            taskid = task_id['task_id']
            trackinfo = await client.collections.track_job([taskid])
            status = trackinfo[taskid]['status']
            if status == 'Completed':
                client.logger.info('Data generated successfully!')
                break
            elif status in ['Failed', 'Cancelled', 'Error']:
                raise RuntimeError(f"Job {taskid} failed with status: {status}")
            else:
                elapsed_time = time.time() - start_time
                client.logger.info(f"Job status: {status} - Elapsed time: {elapsed_time:.1f}s")
                await asyncio.sleep(poll_interval)
                
                
        except KeyboardInterrupt:
            client.logger.info(f"\nInterrupted! Job {taskid} is still running in the background.")
            raise
        except Exception as e:
            client.logger.info(f"\nError tracking job: {e}")
            raise

    os.unlink(tempreqname)

    combine_result = await client.collections.combine_tiles(body["name"], body["overwrite"], body["output"], max_file_size_mb=max_file_size_mb)
    combine_task_id = combine_result.get("task_id")

    combine_start_time = time.time()
    client.logger.info(f"Tracking file generation job {combine_task_id}...")
    while True:
        try:
            trackinfo = await client.collections.track_job([combine_task_id])
            if body["output"] == "netcdf":
                download_file_name = trackinfo[combine_task_id]['folder'] + '.nc'
            elif body["output"] == "geotiff":
                download_file_name = trackinfo[combine_task_id]['folder'] + '.tif'
            bucket = trackinfo[combine_task_id]['bucket']
            combine_status = trackinfo[combine_task_id]['status']
            if combine_status == 'Completed':
                client.logger.info('File/s generated successfully!')
                break
            elif combine_status in ['Failed', 'Cancelled', 'Error']:
                raise RuntimeError(f"File generation job {combine_task_id} failed with status: {combine_status}")
            else:
                elapsed_time = time.time() - combine_start_time
                client.logger.info(f"File generation job status: {combine_status} - Elapsed time: {elapsed_time:.1f}s")
                time.sleep(poll_interval)
        except KeyboardInterrupt:
            client.logger.info(f"\nInterrupted! File generation job {combine_task_id} is still running in the background.")
            raise
        except Exception as e:
            client.logger.info(f"\nError tracking file generation job: {e}")
            raise

    if download_path:
        await client.collections.download_file(
            job_name=body["name"],
            bucket=bucket,
            file_type='processed',
            folder='file-gen',
            page_size=100,
            output_path=download_path,
        )
    else:
        path = f"{body['name']}/outputs/merged/{download_file_name}"
        client.logger.info(f"Dataset file/s is available at {path}")

    return {"generation_task_id": task_id, "combine_task_id": combine_task_id}
